###############################################################################
##
## Copyright (C) 2014-2016, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################



from PyQt5 import QtCore, QtGui, QtWidgets


from vistrails.core import debug
from vistrails.core.configuration import get_vistrails_configuration, \
    get_vistrails_persistent_configuration
from vistrails.core.modules.vistrails_module import ModuleSuspended
from vistrails.core.vistrail.job import module_name
from vistrails.gui import theme
from vistrails.gui.common_widgets import QDockPushButton
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface

import vistrails.core.system
import vistrails.gui.utils

refresh_states = [('Off', 0), ('10 sec', 10),
                  ('1 min', 60), ('10 min', 600),
                  ('1 hour', 3600)]


class QNumberValidator(QtGui.QIntValidator):
    """Variant of QIntValidator that rejects Intermediate values.

    Intermediate strings are strings that could be the left part of an
    Acceptable string.
    """
    def validate(self, input, pos):
        result = QtGui.QIntValidator.validate(self, input, pos)
        if len(input) and result[0] == QtGui.QIntValidator.Intermediate:
            return (QtGui.QIntValidator.Invalid, pos)
        return result


class QJobTree(QtWidgets.QTreeWidget):
    def __init__(self, parent=None):
        QtWidgets.QTreeWidget.__init__(self, parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setColumnCount(2)
        self.header().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        self.header().setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
        self.header().close()
        self.setExpandsOnDoubleClick(False)
        self.controller = None

    def contextMenuEvent(self, event):
        item = self.itemAt(event.pos())
        menu = QtWidgets.QMenu(self)
        if item and isinstance(item, QJobItem):
            act = QtWidgets.QAction("&Check", self)
            act.triggered.connect(lambda :self.parent().check_jobs(item))
            menu.addAction(act)

            act = QtWidgets.QAction("&Delete", self)
            act.triggered.connect(lambda :self.parent().delete_item(item))
            menu.addAction(act)

            act = QtWidgets.QAction("View Standard &Output", self)
            act.setStatusTip("View Standard Output in new window")
            act.triggered.connect(item.stdout)
            menu.addAction(act)

            act = QtWidgets.QAction("View Standard &Error", self)
            act.triggered.connect(item.stderr)
            menu.addAction(act)

            menu.exec_(event.globalPos())

        if item and isinstance(item, QWorkflowItem):

            act = QtWidgets.QAction("&Check", self)
            act.triggered.connect(lambda :self.parent().check_jobs(item))
            menu.addAction(act)

            if not item.paused and not item.workflowFinished:
                act = QtWidgets.QAction("&Pause", self)
                act.triggered.connect(item.pause)
                menu.addAction(act)

            act = QtWidgets.QAction("Delete", self)
            act.triggered.connect(lambda :self.parent().delete_item(item))

            menu.addAction(act)
            menu.exec_(event.globalPos())

class QJobView(QtWidgets.QWidget, QVistrailsPaletteInterface):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.timer_id = None
        self.updating_now = False
        self.widgets = {}

        self.layout = QtWidgets.QVBoxLayout()

        buttonsLayout = QtWidgets.QHBoxLayout()
        run_now = QDockPushButton("Check selected")
        run_now.setToolTip("Check selected job now")
        run_now.clicked.connect(self.check_selected_job)
        buttonsLayout.addWidget(run_now)
        run_all = QDockPushButton("Check all")
        run_all.setToolTip("Check all jobs now")
        run_all.clicked.connect(self.timerEvent)
        buttonsLayout.addWidget(run_all)
        label = QtWidgets.QLabel('Refresh interval (seconds):')
        buttonsLayout.addWidget(label)

        self.interval = QtWidgets.QComboBox()
        for text, seconds in refresh_states:
            self.interval.addItem(text, seconds)
        self.interval.setCompleter(None)
        self.interval.setEditable(True)
        self.interval.editTextChanged.connect(self.set_refresh)
        self.interval.setValidator(QNumberValidator())
        conf = get_vistrails_configuration()
        self.interval.setEditText(str(conf.jobCheckInterval))
        buttonsLayout.addWidget(self.interval)

        self.autorun = QtWidgets.QCheckBox("Automatic re-execution")
        self.autorun.setToolTip("Automatically re-execute workflow when jobs "
                                "complete")
        self.autorun.toggled.connect(self.autorunToggled)
        self.autorun.setChecked(conf.jobAutorun)
        buttonsLayout.addWidget(self.autorun)

        buttonsLayout.addStretch(1)
        self.layout.addLayout(buttonsLayout)

        self.jobView = QJobTree()
        self.jobView.itemDoubleClicked.connect(self.item_clicked)
        self.layout.addWidget(self.jobView)

        self.setLayout(self.layout)

        self.setWindowTitle('Running Jobs')
        self.resize(QtCore.QSize(800, 600))

    def set_controller(self, controller):
        # check if a controller has been closed
        from vistrails.gui.vistrails_window import _app
        controllers = [view.controller for view in _app.getAllViews()]
        for c in list(self.widgets.keys()):
            if c not in controllers:
                self.jobView.takeTopLevelItem(self.jobView.indexOfTopLevelItem(self.widgets[c]))
                del self.widgets[c]

        if not controller:
            return

        # check if controller has been added
        if controller not in self.widgets and controller in controllers:
            item = QVistrailItem(controller)
            self.jobView.addTopLevelItem(item)
            self.jobView.expandAll()
            self.widgets[controller] = item
            if item.childCount() > 0:
                self.set_visible(True)

    def autorunToggled(self, value):
        get_vistrails_configuration().jobAutorun = value
        get_vistrails_persistent_configuration().jobAutorun = value

    def set_refresh(self, refresh=0):
        """Changes the timer time.

        Called when the QComboBox self.interval changes. Updates the
        configuration and restarts the timer.
        """
        self.updating_now = True
        refresh = str(refresh) if refresh else '0'
        if refresh in dict(refresh_states):
            refresh = dict(refresh_states)[refresh]
            self.interval.setEditText(str(refresh))
        else:
            refresh = int(refresh)
        if refresh:
            if self.timer_id is not None:
                self.killTimer(self.timer_id)
            self.timer_id = self.startTimer(refresh*1000)
        else:
            if self.timer_id:
                self.killTimer(self.timer_id)
                self.timer_id = None
        get_vistrails_configuration().jobCheckInterval = refresh
        get_vistrails_persistent_configuration().jobCheckInterval = refresh
        self.updating_now = False

    def update_job(self, job, force=True):
        """ Checks specified job

            force: bool - True means we should ask user to resume jobs
            that has been paused
        """
        if isinstance(job, QJobItem):
            vistrail_item = job.vistrail()
            workflow_item = job.workflow()
        elif isinstance(job, QWorkflowItem):
            vistrail_item = job.parent()
            workflow_item = job
            job = None
        else:
            for workflow_item in list(job.workflowItems.values()):
                self.update_job(workflow_item, force)
            return
        jm = vistrail_item.jobMonitor
        workflow = workflow_item.workflow
        if workflow_item.workflowFinished:
            return
        # re-execute if no handle is set
        if not workflow_item.has_handle:
            if workflow_item.paused:
                if force:
                    workflow_item.resume()
                else:
                    return
            try:
                int(workflow.version)
            except ValueError:
                if (workflow.version.startswith("Parameter Exploration") or
                    workflow.version.startswith("Mashup")):
                    QtWidgets.QMessageBox.question(self, "Running job(s) found",
                        'Running jobs in "%s" are not yet monitored. Rerun it to start monitoring.' %
                                             workflow_item.text(0),
                        QtWidgets.QMessageBox.Ok)
                    # do not notify again
                    workflow_item.pause()
                return
            # Ask user to re-execute workflow
            ret = QtWidgets.QMessageBox.question(self, "Running job(s) found",
                    'Running jobs in workflow "%s" are not yet monitored. Load and check now?' %
                                             workflow_item.text(0),
                    QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
            if ret == QtWidgets.QMessageBox.No:
                workflow_item.pause()
                return
            # restart job and execute
            jm.startWorkflow(workflow)
            self.updating_now = False
            workflow_item.execute()
            self.updating_now = True
            return

        job_items = list(workflow_item.jobs.values()) if job is None else [job]
        for job_item in job_items:
            if job_item.job.finished or job_item.job.ready:
                continue
            try:
                # call monitor
                if jm.isDone(job_item.handle):
                    job_item.job.ready = True
            except Exception as e:
                debug.critical("Error checking job %s: %s" %
                               (workflow_item.text(0), e))
        if workflow_item.updateJobs():
            QJobView.instance().set_visible(True)

        if workflow_item.workflowFinished:
            try:
                int(workflow.version)
            except ValueError:
                if workflow.version.startswith("Parameter Exploration"):
                    # this is too annoying
                    #QtGui.QMessageBox.information(self, "Job Ready",
                    #    'Pending Jobs in "%s" have finished, '
                    #    'execute it again to get results' % workflow_item.text(0),
                    #    QtGui.QMessageBox.Ok)
                    # TODO: Only notify user when all cells have finished
                    pass
                elif workflow.version.startswith("Mashup"):
                    QtWidgets.QMessageBox.information(self, "Job Ready",
                        'Pending Jobs in "%s" have finished, '
                        'execute it again to get results' % workflow_item.text(0),
                        QtWidgets.QMessageBox.Ok)
                return
            if self.autorun.isChecked():
                jm.startWorkflow(workflow)
                self.updating_now = False
                workflow_item.execute()
                self.updating_now = True
                return
            ret = QtWidgets.QMessageBox.information(self, "Job Ready",
                    'Pending Jobs in workflow "%s" have finished, '
                    'continue execution now?' % workflow_item.text(0),
                    QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Cancel)
            if ret == QtWidgets.QMessageBox.Ok:
                jm.startWorkflow(workflow)
                self.updating_now = False
                workflow_item.execute()
                self.updating_now = True

    def check_jobs(self, job=None):
        if self.updating_now:
            return
        self.updating_now = True
        try:
            if job is None:
                for i in range(self.jobView.topLevelItemCount()):
                    vistrail_item = self.jobView.topLevelItem(i)
                    self.update_job(vistrail_item, force=False)
            else:
                self.update_job(job)
        finally:
            self.updating_now = False

    def check_selected_job(self):
        items = self.jobView.selectedItems()
        if len(items) != 1:
            return
        self.check_jobs(items[0])

    def timerEvent(self, id=None):
        self.check_jobs()

    def keyPressEvent(self, event):
        if event.key() in [QtCore.Qt.Key_Delete, QtCore.Qt.Key_Backspace]:
            for item in self.jobView.selectedItems():
                self.delete_item(item)
        else:
            QtWidgets.QWidget.keyPressEvent(self, event)

    def item_clicked(self, item):
        """Item activated.
        """
        if isinstance(item, QWorkflowItem):
            item.goto()

    def delete_item(self, item):
        if isinstance(item, QWorkflowItem):
            item.parent().controller.set_changed(True)
            item.parent().jobMonitor.deleteWorkflow(item.workflow.id)
        elif isinstance(item, QJobItem):
            # find parent workflow
            item.vistrail().controller.set_changed(True)
            item.vistrail().jobMonitor.deleteJob(item.job.id)

class QJobProgressDialog(QtWidgets.QDialog):
    def __init__(self, job_name, job_start, parent=None, value=0, max_value=100):
        super(QJobProgressDialog, self).__init__(parent)
        f = self.windowFlags()
        f |= QtCore.Qt.CustomizeWindowHint
        f &= ~QtCore.Qt.WindowCloseButtonHint
        self.job_name = job_name
        self.job_start = job_start
        self.was_cancelled = False
        self.was_checked = False
        self.progressbar = QtWidgets.QProgressBar()
        self.progressbar.setMinimum(0)
        self.progressbar.setMaximum(max_value)
        self.progressbar.setValue(value)
        self.cancel_button = QtWidgets.QPushButton('&Cancel')
        self.check_now_button = QtWidgets.QPushButton('Check &Now')
        self.cancel_button.clicked.connect(lambda :setattr(self,
                                                           'was_cancelled',
                                                           True))
        self.check_now_button.clicked.connect(lambda :setattr(self,
                                                              'was_checked',
                                                              True))
        self.label = QtWidgets.QLabel()
        self.updateLabel()
        main_layout = QtWidgets.QGridLayout()
        main_layout.addWidget(self.label, 0, 0, 2, 2)
        main_layout.addWidget(self.progressbar, 2, 0, 1, 2)
        main_layout.addWidget(self.check_now_button, 3, 0)
        main_layout.addWidget(self.cancel_button, 3, 1)
        self.setLayout(main_layout)
        self.setWindowTitle('Running Job')
        self.setWindowModality(QtCore.Qt.WindowModal)

    def updateLabel(self):
        labelText = (("Running external job %s\n"
                               "Started %s\n"
                               "Last checked %s\n"
                               "Press Check Now to check or Cancel to suspend")
                               % (self.job_name,
                                  self.job_start,
                                  QtCore.QDateTime.currentDateTime().toString(
                                      QtCore.Qt.ISODate)))
        self.label.setText(labelText)


class QVistrailItem(QtWidgets.QTreeWidgetItem):
    """A vistrail with running workflows.

    This top-level item can have QWorkflowItem's as children.
    """
    def __init__(self, controller, parent=None):
        self.controller = controller
        self.jobMonitor = controller.jobMonitor
        self.jobMonitor.setCallback(self)
        self.locator = controller.vistrail.locator
        QtWidgets.QTreeWidgetItem.__init__(self, parent,
                                       [self.locator.short_name, ''])
        self.setIcon(0, theme.get_current_theme().HISTORY_ICON)
        self.setToolTip(0, self.locator.to_url())
        self.workflowItems = {}
        self.load_running_jobs()

    def load_running_jobs(self):
        """Loads the current jobs from the JSON file.
        """
        workflows = self.jobMonitor.workflows
        # update gui
        for workflow in workflows.values():
            if workflow.id not in self.workflowItems:
                workflow_item = QWorkflowItem(workflow, self)
                self.workflowItems[workflow.id] = workflow_item
                for job in workflow.jobs.values():
                    if job.id not in workflow_item.jobs:
                        workflow_item.jobs[job.id] = QJobItem(job, workflow_item)
                        workflow_item.updateJobs()

    def startWorkflow(self, workflow):
        """Empty callback.
        """

    def addJob(self, job):
        """ addJob(self, job: job.Module) -> None
        Callback, adds or updates a job in the interface.
        """

        workflow = self.jobMonitor.currentWorkflow()
        if workflow.id not in self.workflowItems:
            workflow_item = QWorkflowItem(workflow, self)
            workflow_item.setExpanded(True)
            self.workflowItems[workflow.id] = workflow_item

        workflow_item = self.workflowItems[workflow.id]
        changed = False
        if job.id not in workflow_item.jobs:
            workflow_item.jobs[job.id] = QJobItem(job, workflow_item)
            changed = True
        if workflow_item.updateJobs() or changed:
            QJobView.instance().set_visible(True)

    def deleteWorkflow(self, id):
        """ deleteWorkflow(id: str) -> None
        Callback, deletes a workflow.
        """
        self.takeChild(self.indexOfChild(self.workflowItems[id]))
        del self.workflowItems[id]

    def deleteJob(self, id):
        """ deleteJob(id: str, parent_id: str) -> None
        Callback, deletes a a single job from all workflows.

        """

        for workflow_item in self.workflowItems.values():
            if id in workflow_item.jobs:
                job_item = workflow_item.jobs[id]
                job_item.parent().takeChild(job_item.parent().indexOfChild(job_item))
                del workflow_item.jobs[id]
                if workflow_item.updateJobs():
                    QJobView.instance().set_visible(True)

    def addJobRec(self, obj, parent_id=None):
        """addJobRec(obj: ModuleSuspended, parent_id: signature)  -> None

           Recursively adds jobs that are executed by other modules like
           Groups and Maps. This is only for display purposes.
        """
        workflow = self.jobMonitor.currentWorkflow()
        workflow_item = self.workflowItems[workflow.id]
        # top down. Base is assumed to have been added already
        base = (workflow_item.intermediates[parent_id] if parent_id is not None
                                                       else workflow_item)
        id = obj.module.signature
        if obj.children:
            # add parent items and their children
            if id not in workflow_item.intermediates:
                parent_item = QParentItem(id, module_name(obj.module), base)
                parent_item.setExpanded(True)
                workflow_item.intermediates[id] = parent_item

            for child in obj.children:
                self.addJobRec(child, id)
        elif obj.module.signature in workflow.jobs:
            # this is an already existing new-style job
            job = workflow_item.jobs[obj.module.signature]
            job.handle = obj.handle
            # need to force takeChild
            base.addChild(job.parent().takeChild(job.parent().indexOfChild(job)))
        elif id in workflow.jobs:
            # this is an already existing old-style job
            job = workflow_item.jobs[id]
            job.handle = obj.handle
            # need to force takeChild
            base.addChild(job.parent().takeChild(job.parent().indexOfChild(job)))

    def finishWorkflow(self, workflow):
        """Callback, updates workflow status.
        """
        workflow = self.jobMonitor.currentWorkflow()
        # untangle parents
        for parent in workflow.parents.values():
            self.addJobRec(parent)

        workflow_item = self.workflowItems.get(workflow.id, None)
        if workflow_item:
            if workflow_item.updateJobs():
                QJobView.instance().set_visible(True)

    def checkJob(self, module, id, handle):
        """ checkJob(module: VistrailsModule, id: str, handle: object)
        Callback, checks if job has completed.
        """
        workflow = self.jobMonitor.currentWorkflow()
        if not workflow:
            if not self.jobMonitor.isDone(handle):
                raise ModuleSuspended(module, 'Job is running',
                                      handle=handle)
        workflow_item = self.workflowItems[workflow.id]
        item = workflow_item.jobs.get(id, None)
        item.setText(0, item.job.name)
        # we should check the status using the JobHandle and show dialog
        # get current view progress bar and hijack it
        item.handle = handle
        workflow = self.jobMonitor.currentWorkflow()
        workflow_item = self.workflowItems.get(workflow.id, None)
        workflow_item.updateJobs()
        conf = get_vistrails_configuration()
        interval = conf.jobCheckInterval
        is_done = self.jobMonitor.isDone(handle)
        old_progress = self.controller.progress
        if (interval and not conf.jobAutorun and
               not old_progress.suspended and not is_done):
            # wait for module to complete
            old_progress.hide()
            progress = QJobProgressDialog(item.job.name,
                                          item.job.start,
                                          old_progress.parent(),
                                          old_progress.value(),
                                          old_progress.maximum())
            progress.show()
            while not is_done:
                dieTime = QtCore.QDateTime.currentDateTime().addSecs(interval)
                while QtCore.QDateTime.currentDateTime() < dieTime and not is_done:
                    QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 100)
                    if progress.was_checked:
                        progress.check_now_button.setText('Checking')
                        progress.check_now_button.setEnabled(False)
                        progress.cancel_button.setEnabled(False)
                        is_done = self.jobMonitor.isDone(handle)
                        progress.check_now_button.setText('Check &Now')
                        progress.check_now_button.setEnabled(True)
                        progress.cancel_button.setEnabled(True)
                        progress.was_checked = False
                        progress.updateLabel()
                    elif progress.was_cancelled:
                        # this does not work, need to create a new progress dialog
                        #old_progress.goOn()
                        new_progress = old_progress.__class__(old_progress.parent())
                        new_progress.setMaximum(old_progress.maximum())
                        new_progress.setValue(old_progress.value())
                        new_progress.setLabelText(old_progress.labelText())
                        new_progress.setMinimumDuration(0)
                        new_progress.suspended = True
                        self.controller.progress = new_progress
                        old_progress.deleteLater()
                        progress.hide()
                        new_progress.show()
                        QtCore.QCoreApplication.processEvents()
                        raise ModuleSuspended(module,
                                   'Interrupted by user, job'
                                   ' is still running', handle=handle)
                is_done = is_done or self.jobMonitor.isDone(handle)
            # is_done!
            new_progress = old_progress.__class__(old_progress.parent())
            new_progress.setMaximum(old_progress.maximum())
            new_progress.setValue(old_progress.value())
            new_progress.setLabelText(old_progress.labelText())
            new_progress.setMinimumDuration(0)
            new_progress.suspended = True
            self.controller.progress = new_progress
            old_progress.deleteLater()
            progress.hide()
            new_progress.show()
            QtCore.QCoreApplication.processEvents()
            return
        if not is_done:
            raise ModuleSuspended(module, 'Job is running', handle=handle)


class QWorkflowItem(QtWidgets.QTreeWidgetItem):
    """A workflow with jobs.

       It can reference a workflow version, a parameter exploration workflow,
       or a mashup

    This item can have child module items.
    """
    def __init__(self, workflow, parent):
        QtWidgets.QTreeWidgetItem.__init__(self, parent, ['', ''])
        self.workflow = workflow
        self.has_handle = True
        # Paused workflows will not be checked by the timer/"Check all" button
        self.paused = False
        self.setIcon(0, theme.get_current_theme().PIPELINE_ICON)
        self.setIcon(1, theme.get_current_theme().JOB_CHECKING)
        self.workflowFinished = False
        self.jobs = {}
        self.intermediates = {}
        self.updateJobs()

    def updateJobs(self):
        """ Updates name and job states
        """
        self.paused = False
        try:
            name = self.parent().controller.get_pipeline_name(
                                                           self.workflow.version)
        except (KeyError, TypeError):
            name = self.workflow.version
        self.setToolTip(0, 'Double-Click to View Pipeline "%s" with id "%s"' %
                           (name, self.workflow.version))
        self.setText(0, name)
        self.setToolTip(1, "Log id: %s" % self.workflow.id)
        changed = False
        self.has_handle = True
        for job in self.jobs.values():
            if job.updateJob():
                changed = True
            if not job.job.finished and not job.handle:
                self.has_handle = False
        count = len(self.jobs)
        finished = sum([job.job.finished or job.job.ready for job in list(self.jobs.values())])
        self.setText(1, "(%s/%s)" % (finished, count))
        self.workflowFinished = (finished == count)
        if self.workflowFinished:
            self.setIcon(1, theme.get_current_theme().JOB_FINISHED)
        elif not self.has_handle:
            self.setIcon(1, theme.get_current_theme().JOB_SCHEDULED)
        else:
            self.setIcon(1, theme.get_current_theme().JOB_CHECKING)
        return changed

    def goto(self):
        """ Shows this pipeline

        """
        try:
            int(self.workflow.version)
        except ValueError:
            # this is not a pipeline id
            return
        from vistrails.gui.vistrails_window import _app
        view = _app.getViewFromLocator(self.parent().controller.locator)
        _app.change_view(view)
        view.version_selected(self.workflow.version, True, double_click=True)
        return view

    def execute(self):
        """ Shows and executes this pipeline
        """
        try:
            int(self.workflow.version)
        except ValueError:
            # this is not a pipeline id
            return
        self.goto().execute()

    def pause(self):
        self.paused = True
        self.setIcon(1, theme.get_current_theme().JOB_SCHEDULED)

    def resume(self):
        self.paused = False
        self.setIcon(1, theme.get_current_theme().JOB_CHECKING)

class QJobItem(QtWidgets.QTreeWidgetItem):
    """A pending job, i.e. a single module that was suspended.

    These will be either under a QWorkflowItem or a QParentItem.
    """
    def __init__(self, job, parent=None):
        QtWidgets.QTreeWidgetItem.__init__(self, parent, [job.name,
                                                      job.description()])
        self.setToolTip(1, job.description())
        self.job = job
        self.handle = None
        self.finished = None
        self.updateJob()

    def updateJob(self):
        changed = self.job.finished != self.finished
        self.finished = self.job.finished
        self.setText(1, self.job.parameters.get('__message__',
                        "Finished" if self.job.finished or self.job.ready else "Running"))
        if self.job.finished or self.job.ready:
            self.setIcon(1, theme.get_current_theme().JOB_FINISHED)
            self.setToolTip(0, "This Job Has Finished")
        elif self.handle:
            self.setIcon(1, theme.get_current_theme().JOB_SCHEDULED)
            self.setToolTip(0, "This Job is Running and Scheduled for Checking")
        else:
            self.setIcon(1, theme.get_current_theme().JOB_CHECKING)
            self.setToolTip(0, "This Job is Running")
        self.setToolTip(1, self.job.id.decode())
        return changed

    def stdout(self):
        if self.handle:
            sp = LogMonitor("Standard Output for " + self.job.name,
                            self.handle)
            sp.exec_()

    def stderr(self):
        if self.handle:
            sp = ErrorMonitor("Standard Output for " + self.job.name,
                              self.handle)
            sp.exec_()

    def workflow(self):
        # find parent workflow
        parent = self.parent()
        while not isinstance(parent, QWorkflowItem):
            parent = parent.parent()
        return parent

    def vistrail(self):
        # find parent vistrail
        return self.workflow().parent()

class QParentItem(QtWidgets.QTreeWidgetItem):
    """A composite job, i.e. a module whose child suspended.
    """
    def __init__(self, id, name, parent=None):
        QtWidgets.QTreeWidgetItem.__init__(self, parent, [name, ''])
        self.id = id
        self.setToolTip(0, self.id.decode())


class LogMonitor(QtWidgets.QDialog):
    """Displays the content of a Job's standard_output().
    """
    def __init__(self, name, handle, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.handle = handle
        self.resize(700, 400)
        self.setWindowTitle(name)

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        self.text = QtWidgets.QTextEdit('')
        self.update_text()
        self.text.setReadOnly(True)
        self.text.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        layout.addWidget(self.text)
        buttonLayout = QtWidgets.QHBoxLayout()

        close = QtWidgets.QPushButton('Close', self)
        close.setFixedWidth(100)
        buttonLayout.addWidget(close)
        close.clicked.connect(self.close)

        update = QtWidgets.QPushButton('Update', self)
        update.setFixedWidth(100)
        buttonLayout.addWidget(update)
        update.clicked.connect(self.update_text)

        layout.addLayout(buttonLayout)

    def update_text(self):
        self.text.setPlainText(self.handle.standard_output())


class ErrorMonitor(LogMonitor):
    """Displays the content of a job's standard_error().
    """
    def update_text(self):
        self.text.setPlainText(self.handle.standard_error())


################################################################################
# Testing


class TestJobMonitor(vistrails.gui.utils.TestVisTrailsGUI):

    @classmethod
    def setUpClass(cls):
        get_vistrails_configuration().jobAutorun = True
        get_vistrails_persistent_configuration().jobAutorun = True
        QJobView.instance().set_refresh()
        cls.filename = (vistrails.core.system.vistrails_root_directory() +
                        '/tests/resources/jobs.vt')

        pm = vistrails.core.packagemanager.get_package_manager()
        if pm.has_package('org.vistrails.vistrails.myjobs'):
            return
        d = {'myjob': 'vistrails.tests.resources.'}
        pm.late_enable_package('myjob', d)

    @classmethod
    def tearDownClass(cls):
        manager = vistrails.core.packagemanager.get_package_manager()
        if manager.has_package('org.vistrails.vistrails.myjobs'):
            manager.late_disable_package('myjob')

    def tearDown(self):
        vistrails.gui.utils.TestVisTrailsGUI.tearDown(self)
        from vistrails.core.interpreter.cached import CachedInterpreter
        CachedInterpreter.flush()

    def testSuspended(self):
        from vistrails import api
        view = api.open_vistrail_from_file(self.filename)
        c = view.controller
        api.select_version('SuspendOnce', c)

        result = c.execute_user_workflow()[0][0]
        # assert suspended
        self.assertEqual(result.errors, {})
        self.assertNotEqual(result.suspended, {})

        result = c.execute_user_workflow()[0][0]
        # assert finished
        self.assertEqual(result.errors, {})
        self.assertEqual(result.suspended, {})

        for i in list(c.jobMonitor.workflows.keys()):
            c.jobMonitor.deleteWorkflow(i)

    def testGroup(self):
        from vistrails import api
        view = api.open_vistrail_from_file(self.filename)
        c = view.controller
        api.select_version('SuspendGroup', c)
        result = c.execute_user_workflow()[0][0]

        # assert suspended
        self.assertEqual(result.errors, {})
        self.assertNotEqual(result.suspended, {})

        for i in list(c.jobMonitor.workflows.keys()):
            c.jobMonitor.deleteWorkflow(i)

    def testMap(self):
        from vistrails import api
        view = api.open_vistrail_from_file(self.filename)
        c = view.controller
        api.select_version('SuspendMap', c)

        result = c.execute_user_workflow()[0][0]
        # assert suspended
        self.assertEqual(result.errors, {})
        self.assertNotEqual(result.suspended, {})

        result = c.execute_user_workflow()[0][0]
        # assert finished
        self.assertEqual(result.errors, {})
        self.assertEqual(result.suspended, {})

        for i in list(c.jobMonitor.workflows.keys()):
            c.jobMonitor.deleteWorkflow(i)

    def testLoop(self):
        from vistrails import api
        view = api.open_vistrail_from_file(self.filename)
        c = view.controller
        api.select_version('SuspendLoop', c)

        result = c.execute_user_workflow()[0][0]
        # assert suspended
        self.assertEqual(result.errors, {})
        self.assertNotEqual(result.suspended, {})

        result = c.execute_user_workflow()[0][0]
        # assert finished
        self.assertEqual(result.errors, {})
        self.assertEqual(result.suspended, {})

        for i in list(c.jobMonitor.workflows.keys()):
            c.jobMonitor.deleteWorkflow(i)

    def testParameterExploration(self):
        from vistrails import api
        view = api.open_vistrail_from_file(self.filename)
        c = view.controller
        api.select_version('SuspendOnce', c)

        pe = c.vistrail.get_named_paramexp('SuspendOnce')
        try:
            c.executeParameterExploration(pe)
        except:
            self.fail("Parameter Exploration with Job failed")

        # Check that we have 2 jobs
        self.assertEqual(len(list(c.jobMonitor.workflows.keys())), 2)
        for i in list(c.jobMonitor.workflows.keys()):
            wf = c.jobMonitor.workflows[i]
            self.assertFalse(wf.completed())

        try:
            c.executeParameterExploration(pe)
        except:
            self.fail("Parameter Exploration with Job failed")

        # Check that the 2 jobs has completed
        for i in list(c.jobMonitor.workflows.keys()):
            wf = c.jobMonitor.workflows[i]
            self.assertTrue(wf.completed())

        for i in list(c.jobMonitor.workflows.keys()):
            c.jobMonitor.deleteWorkflow(i)

    def testMashup(self):
        from vistrails import api
        view = api.open_vistrail_from_file(self.filename)
        c = view.controller

        c.flush_delayed_actions()
        id = "80f58f50-57b1-11e5-a1da-000c2960b7d8"
        mashup = view.get_mashup_from_mashuptrail_id(id, "SuspendOnce")
        self.assert_(mashup)

        view.open_mashup(mashup)
        self.assertEqual(len(list(c.jobMonitor.workflows.keys())), 1)
        self.assertFalse(list(c.jobMonitor.workflows.values())[0].completed())

        # close associated mashup apps
        from vistrails.gui.version_prop import QVersionProp
        version_prop = QVersionProp.instance()
        for app in list(version_prop.versionMashups.apps.values()):
            app.close()

        view.open_mashup(mashup)
        self.assertEqual(len(list(c.jobMonitor.workflows.keys())), 1)
        self.assertTrue(list(c.jobMonitor.workflows.values())[0].completed())

        for i in list(c.jobMonitor.workflows.keys()):
            c.jobMonitor.deleteWorkflow(i)



from vistrails.core.modules.vistrails_module import Module


class C(Module):
    _output_ports = [('result', 'basic:String')]

    def compute(self):
        self.set_output('result', 'C')


_modules = [C]



#!/usr/bin/env python
###########################################################
# Builds Source Release and uploads to Sourceforge        #
# Author: Daniel S. Rees (Modified by Tommy Ellqvist)     #
#                                                         #
# Description:                                            #
#     Creates and archives a tarball of a git             #
#     repository and uploads it to Sourceforge.           #
#     Intended for use as a scheduled task or             #
#     run by a buildbot.                                  #
#                                                         #
# Instructions:                                           #
#     Check out a clean version from git, then run        #
#     dist/common/prepare_release.py to update hash.      #
#     Running this script will create a tarball in the    #
#     root directory with the files in EXPORT_PATHS.      #
#     Specify sourceforge username and path to private    #
#     key to upload to sourceforge.                       #
#                                                         #
# Requires 'scp' to be installed and accessible from a    #
# default environment (use Cygwin if on Windows).         #
###########################################################

import os
import sys
import tarfile
import subprocess

####################
###### Config ######
####################

# Paths of files and/or directories to be included in the exported repository (relative to export dir)
EXPORT_PATHS = "vistrails doc examples extensions scripts CHANGELOG LICENSE".split()

# VisTrails Release Version
VT_VERSION = '2.x'

# Branch to be used to build release
VT_BRANCH = 'master'

# Hash used in the release
VT_HASH = '0fddb5bfe72c'

# Distribution Tarball name (Do not add ".tar.gz")
#TARBALL_NAME = "vistrails-src-%s-%s" % (VT_VERSION, VT_HASH)
TARBALL_NAME = "vistrails-src"

# Tarball file name
TARBALL_FILENAME = TARBALL_NAME + ".tar.gz"

# Flag determines if tarball is uploaded (Can be set manually here or conditionally in last_minute_changes())
SF_DO_UPLOAD = False

# Sourceforge User Name
SF_USERNAME = "CHANGEME"

# Sourceforge Project Name (unix name)
SF_PROJECT = "vistrails"

# Sourceforge target directory for upload (relative path from project root)
SF_DIRNAME = "vistrails/v%s"%VT_VERSION

# Path to private key for secure connection to Sourceforge (absolute path)
SF_PRIVKEY_PATH = "/path/to/private/key/CHANGEME"

# Number of upload attempts before giving up (Setting this to 0 guarantees that no upload will occur)
SF_UPLOAD_ATTEMPTS = 3

# Directory in which tarball archives are stored
ARCHIVE_DIR = "archive"


#########################################
###### Read command-line arguments ######
#########################################

# If called with 2 arguments, get [sf_username, sf_privkey_path] and upload so sourceforge
if len(sys.argv)==3:
    SF_DO_UPLOAD = True
    SF_USERNAME = sys.argv[1]
    SF_PRIVKEY_PATH = sys.argv[2]

    TARBALL_NAME = "vistrails-src-nightly"
    TARBALL_FILENAME = TARBALL_NAME + ".tar.gz"
    SF_DIRNAME = "vistrails/nightly"

#####################
###### Globals ######
#####################

# Sourceforge upload command
SF_UPLOAD_CMD = "scp -v -i %s %s %s,%s@frs.sourceforge.net:/home/frs/project/%s/%s/%s/%s" % (
                    SF_PRIVKEY_PATH, TARBALL_FILENAME, SF_USERNAME, SF_PROJECT, SF_PROJECT[0],
                    SF_PROJECT[0:2], SF_PROJECT, SF_DIRNAME)

if __name__ == "__main__":
    os.chdir(os.path.join(os.path.dirname(__file__), '..', '..', '..'))

    # prepare release
    proc = subprocess.Popen(['scripts/dist/common/prepare_release.py'],
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    proc.wait()
    if proc.returncode != 0:
        print "ERROR: preparing release."
        if proc.stdout:
            print proc.stdout.readlines()
        sys.exit(1)

    print "Creating tarball: '%s' ..." % TARBALL_FILENAME
    tarball = None
    try:
        tarball = tarfile.open(TARBALL_FILENAME, "w:gz")
        ignore = [os.path.join('scripts', 'dist'),
                  os.path.join('doc', 'dist')]

        def filter_(member):
            if any(member.name.startswith(ignored) for ignored in ignore):
                return None
            return member

        for fname in EXPORT_PATHS:
            tarball.add(fname, filter=filter_)
    except:
        print "ERROR: Failed to create tarball"
        raise
    finally:
        if tarball is not None:
            tarball.close()

    if SF_DO_UPLOAD and SF_UPLOAD_ATTEMPTS > 0:
        # Upload to sourceforge
        uploaded = False
        upload_attempts = 0
        while not uploaded and upload_attempts < SF_UPLOAD_ATTEMPTS:
            upload_attempts += 1
            print "Uploading tarball to Sourceforge (attempt %d of %d): '%s' ..." % (upload_attempts, SF_UPLOAD_ATTEMPTS, SF_UPLOAD_CMD)
            try:
                upload_proc = subprocess.Popen(SF_UPLOAD_CMD, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                upload_log = upload_proc.communicate()[0]
                # Indent upload log and print
                upload_log = "\n".join(['    ' + line for line in upload_log.splitlines()])
                print "Sourceforge Upload Log (attempt %d of %d):\n%s" % (upload_attempts, SF_UPLOAD_ATTEMPTS, upload_log)
                if upload_proc.returncode != 0:
                    raise Exception("Tarball upload (attempt %d of %d) failed with return code: %s" % (upload_attempts, SF_UPLOAD_ATTEMPTS, upload_proc.returncode))
                else:
                    print "Tarball upload completed on attempt %d of %d." % (upload_attempts, SF_UPLOAD_ATTEMPTS)
                    uploaded = True
            except Exception, e:
                if upload_attempts == SF_UPLOAD_ATTEMPTS:
                    print "ERROR: Could not upload to sourceforge"
                else:
                    print e.args[0]
        print "tarball uploaded to Sourceforge, deleting local tarball"
        if os.path.isfile(TARBALL_FILENAME):
            os.remove(TARBALL_FILENAME)

    print "Completed Source Build"



###############################################################################
##
## Copyright (C) 2014-2016, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""The package manager takes care of everything that has got to do
with handling packages, from setting paths to adding new packages
to checking dependencies to initializing them."""


import copy
from distutils.version import LooseVersion
import inspect
import itertools
import os
import sys
import warnings

from vistrails.core import debug, get_vistrails_application, system
from vistrails.core.configuration import ConfigurationObject, \
    get_vistrails_configuration
import vistrails.core.data_structures.graph
from vistrails.core.modules.module_registry import MissingPackage, \
    MissingPackageVersion
from vistrails.core.modules.package import Package
from vistrails.core.requirements import MissingRequirement
from vistrails.core.utils import VistrailsInternalError, \
    versions_increasing, VistrailsDeprecation
import vistrails.packages

##############################################################################


global _package_manager
_package_manager = None

class PackageManager(object):
    """Discovers and loads packages of VisTrails modules.

    Signals:
    * add_package_menu_signal is emitted with a tuple containing the package
      identifier, package name and the menu item
    * remove_package_menu_signal is emitted with the package identifier
    * package_error_message_signal is emitted with the package identifier,
      package name and the error message
    * reloading_package_signal is emitted when a package reload has disabled
      the packages, but has not yet enabled them
    """

    class DependencyCycle(Exception):
        """There is a cycle in the dependencies between packages.
        """
        def __init__(self, p1, p2):
            self._package_1 = p1
            self._package_2 = p2
        def __str__(self):
            return ("Packages '%s' and '%s' have cyclic dependencies" %
                    (self._package_1,
                     self._package_2))

    class PackageInternalError(Exception):
        """An error was raised by the package while it was loading.
        """
        def __init__(self, n, d):
            self._package_name = n
            self._description = d
        def __str__(self):
            return "Package '%s' has a bug: %s" % (self._package_name,
                                                   self._description)

    def import_packages_module(self):
        """Imports the 'vistrails.packages' package.

        This might need to manipulate the Python path to find it.
        """
        if self._packages is not None:
            return self._packages
        # Imports standard packages directory
        conf = self._startup.temp_configuration
        old_sys_path = copy.copy(sys.path)
        if conf.check('packageDirectory'):
            sys.path.insert(0, conf.packageDirectory)
        try:
            import vistrails.packages
        except ImportError:
            debug.critical('ImportError: "packages" sys.path: %s' % sys.path)
            raise
        finally:
            sys.path = old_sys_path
        self._packages = vistrails.packages
        return vistrails.packages

    def import_user_packages_module(self):
        """Imports the 'userspackages' package.

        This will need to manipulate the Python path to find it.
        """
        if self._userpackages is not None:
            return self._userpackages
        # Imports user packages directory
        old_sys_path = copy.copy(sys.path)
        userPackageDir = system.get_vistrails_directory('userPackageDir')
        if userPackageDir is not None:
            sys.path.insert(0, os.path.join(userPackageDir, os.path.pardir))
            try:
                import userpackages
            except ImportError:
                debug.critical('ImportError: "userpackages" sys.path: %s' %
                               sys.path)
                raise
            finally:
                sys.path = old_sys_path
            os.environ['VISTRAILS_USERPACKAGES_DIR'] = userPackageDir
            self._userpackages = userpackages
            return userpackages
        # possible that we don't have userPackageDir set!
        return None

    def __init__(self, registry, startup):
        global _package_manager
        if _package_manager:
            m = "Package manager can only be constructed once."
            raise VistrailsInternalError(m)
        _package_manager = self

        self._registry = registry
        self._startup = startup

        # Contains packages that have not yet been enabled, but exist on the
        # filesystem
        self._available_packages = {} # codepath: str -> Package
        # These other lists contain enabled packages
        self._package_list = {} # codepath: str -> Package
        self._package_versions = {} # identifier: str -> version -> Package
        self._old_identifier_map = {} # old_id: str -> new_id: str
        self._dependency_graph = vistrails.core.data_structures.graph.Graph()
        self._default_prefix_dict = \
                                {'basic_modules': 'vistrails.core.modules.',
                                 'abstraction': 'vistrails.core.modules.'}

        self._userpackages = None
        self._packages = None
        self._abstraction_pkg = None
        self._currently_importing_package = None

        # Setup a global __import__ hook that calls Package#import_override()
        # for all imports executed from that package
        import builtins
        self._orig_import = builtins.__import__
        builtins.__import__ = self._import_override

        # Compute the list of available packages, _available_packages
        self.build_available_package_names_list()

        if get_vistrails_configuration().loadPackages:
            for pkg in self._startup.enabled_packages.values():
                self.add_package(pkg.name, prefix=pkg.prefix)
        else:
            try:
                basic_pkg = self._startup.enabled_packages['basic_modules']
            except KeyError:
                pass
            else:
                self.add_package(basic_pkg.name, prefix=basic_pkg.prefix)

            try:
                abs_pkg = self._startup.enabled_packages['abstraction']
            except KeyError:
                pass
            else:
                self.add_package(abs_pkg.name, prefix=abs_pkg.prefix)

    def _import_override(self,
                         name, globals={}, locals={}, fromlist=[], level=0):
        """Overridden __import__ function.

        This replaces the builtin __import__ function globally so that we can
        track imports done from a package. This is recorded in the Package so
        that reloading the package actually reloads all dependent code.
        """
        # Get the caller module, using globals (like the original __import
        # does)
        try:
            if globals is None:
                raise KeyError
            module = globals['__name__']
        except (KeyError, TypeError):
            # Another method of getting the caller module, using the stack
            caller = inspect.currentframe().f_back
            module = inspect.getmodule(caller)
            # Some frames might not be associated to a module, because of the
            # use of exec for instance; we just skip these until we reach a
            # valid one
            while module is None:
                caller = caller.f_back
                if caller is None:
                    break
                module = inspect.getmodule(caller)
            if module:
                module = module.__name__

        # Get the Package from the module name
        if module:
            importing_pkg = None
            current = self._currently_importing_package
            if (current is not None and
                    current.prefix and
                    module.startswith(current.prefix + current.codepath)):
                importing_pkg = current
            else:
                for pkg in itertools.chain(
                        self._package_list.values(),
                        self._available_packages.values()):
                    if (pkg.prefix is not None and
                            module.startswith(pkg.prefix + pkg.codepath)):
                        importing_pkg = pkg
                        break

            # If we are importing directly from a package
            if importing_pkg is not None:
                old_current = self._currently_importing_package
                self._currently_importing_package = importing_pkg
                result = importing_pkg.import_override(
                        self._orig_import,
                        name, globals, locals, fromlist, level,
                        package_importing_directly=True)
                self._currently_importing_package = old_current
                return result
            # If we are doing it indirectly (from other stuff imported from a
            # package)
            elif self._currently_importing_package is not None:
                return self._currently_importing_package.import_override(
                        self._orig_import,
                        name, globals, locals, fromlist, level,
                        package_importing_directly=False)

        # Else, this is not from a package
        return self._orig_import(name, globals, locals, fromlist, level)

    def finalize_packages(self):
        """Finalizes all initialized packages.
        """
        for package in self._package_list.values():
            package.finalize()
        self._package_list = {}
        self._package_versions = {}
        self._old_identifier_map = {}
        global _package_manager
        _package_manager = None

    def get_available_package(self, codepath, prefix=None):
        try:
            pkg = self._available_packages[codepath]
        except KeyError:
            pkg = self._registry.create_package(codepath, prefix=prefix)
            self._available_packages[codepath] = pkg
        pkg.persistent_configuration = \
                                self._startup.get_pkg_configuration(codepath)
        return pkg

    def add_package(self, codepath, add_to_package_list=True, prefix=None):
        """Adds a new package to the manager. This does not initialize it.

        You will have to call initialize_packages() to initialize loaded but
        uninitialized packages.
        """
        package = self.get_available_package(codepath)
        if add_to_package_list:
            self.add_to_package_list(codepath, package, prefix)
        return package

    def add_to_package_list(self, codepath, package, prefix=None):
        self._available_packages[codepath] = package
        self._package_list[codepath] = package
        if prefix is not None:
            self._default_prefix_dict[codepath] = prefix

    def remove_old_identifiers(self, identifier):
        # remove refs in old_identifier_map
        old_ids = []
        for old_id, cur_id in self._old_identifier_map.items():
            if cur_id == identifier:
                old_ids.append(old_id)
        for old_id in old_ids:
            del self._old_identifier_map[old_id]

    def remove_package(self, codepath):
        """Removes a package from the system.

        Use late_disable_package() in application code so that the
        configuration is updated.
        """
        pkg = self._package_list[codepath]

        from vistrails.core.interpreter.cached import CachedInterpreter
        CachedInterpreter.clear_package(pkg.identifier)

        self._dependency_graph.delete_vertex(pkg.identifier)
        del self._package_versions[pkg.identifier][pkg.version]
        if len(self._package_versions[pkg.identifier]) == 0:
            del self._package_versions[pkg.identifier]
        self.remove_old_identifiers(pkg.identifier)
        self.remove_menu_items(pkg)
        pkg.finalize()
        del self._package_list[codepath]
        self._registry.remove_package(pkg)
        app = get_vistrails_application()
        app.send_notification("package_removed", codepath)

    def has_package(self, identifier, version=None):
        """Returns true if given package identifier is present.
        """

        # check if it's an old identifier
        identifier = self._old_identifier_map.get(identifier, identifier)
        if identifier in self._package_versions:
            return (version is None or
                    version in self._package_versions[identifier])
        return False

    def look_at_available_package(self, codepath):
        """Returns a Package object for an uninitialized package.

        This does NOT initialize it.
        """
        return self.get_available_package(codepath)

    def get_package(self, identifier, version=None):
        # check if it's an old identifier
        identifier = self._old_identifier_map.get(identifier, identifier)
        try:
            package_versions = self._package_versions[identifier]
            if version is not None:
                return package_versions[version]
        except KeyError:
            # dynamic packages are only registered in the registry
            try:
                return self._registry.get_package_by_name(identifier, version)
            except MissingPackageVersion:
                return self._registry.get_package_by_name(identifier)

        max_version = '0'
        max_pkg = None
        for version, pkg in package_versions.items():
            if versions_increasing(max_version, version):
                max_version = version
                max_pkg = pkg
        return max_pkg

    def get_package_by_codepath(self, codepath):
        """Returns a package with given codepath if it is initialized.

        :raises MissingPackage: if the package is not loaded or initialized.
        """
        if codepath not in self._package_list:
            raise MissingPackage(codepath)
        else:
            return self._package_list[codepath]

    def get_package_by_identifier(self, identifier):
        """Deprecated, use get_package() instead.
        """
        warnings.warn(
                "You should use get_package instead of "
                "get_package_by_identifier",
                VistrailsDeprecation,
                stacklevel=2)
        return self.get_package(identifier)

    def get_package_configuration(self, codepath):
        """Returns the configuration object for the package, or None.

        This will only return None if the package exists but has no
        configuration object.
        :raises MissingPackage: if the package is not loaded or initialized.
        """
        pkg = self.get_package_by_codepath(codepath)

        if not hasattr(pkg.module, 'configuration'):
            return None
        else:
            c = pkg.module.configuration
            if not isinstance(c, ConfigurationObject):
                d = "'configuration' attribute should be a ConfigurationObject"
                raise self.PackageInternalError(codepath, d)
            return c

    def check_dependencies(self, package, deps):
        # want to check that necessary version also exists, if specified
        missing_deps = []
        for dep in deps:
            min_version = None
            max_version = None
            if isinstance(dep, tuple):
                identifier = dep[0]
                if len(dep) > 1:
                    min_version = dep[1]
                    if len(dep) > 2:
                        max_version = dep[2]
            else:
                identifier = dep

            # check if it's an old identifier
            identifier = self._old_identifier_map.get(identifier, identifier)
            if identifier not in self._package_versions:
                missing_deps.append((identifier, None, None))
            else:
                if min_version is None and max_version is None:
                    continue
                found_version = False
                for version, pkg in \
                        self._package_versions[identifier].items():
                    if ((min_version is None or
                         versions_increasing(min_version, version)) and
                        (max_version is None or
                         versions_increasing(version, max_version))):
                        found_version = True
                if not found_version:
                    missing_deps.append((identifier, min_version, max_version))

        if len(missing_deps) > 0:
            raise Package.MissingDependency(package, missing_deps)
        return True

    def add_dependencies(self, package):
        """Register all the dependencies of a package contains.
        """
        deps = package.dependencies()
        # FIXME don't hardcode this
        from vistrails.core.modules.basic_modules import identifier as basic_pkg
        if package.identifier != basic_pkg:
            deps.append(basic_pkg)

        self.check_dependencies(package, deps)

        for dep in deps:
            if isinstance(dep, tuple):
                dep_name = dep[0]
            else:
                dep_name = dep
            dep_name = self.get_package(dep_name).identifier

            if not self._dependency_graph.has_edge(package.identifier,
                                                   dep_name):
                self._dependency_graph.add_edge(package.identifier, dep_name)

    def late_enable_package(self, codepath, prefix_dictionary={},
                            needs_add=True):
        """Enables a package 'late', i.e. after VisTrails initialization.

        Note that all the dependencies need to already be enabled.
        """
        if needs_add:
            if codepath in self._package_list:
                msg = 'duplicate package identifier: %s' % codepath
                raise VistrailsInternalError(msg)
            self.add_package(codepath)
        app = get_vistrails_application()
        pkg = self.get_package_by_codepath(codepath)
        try:
            pkg.load(prefix_dictionary.get(pkg.codepath, None))
            # pkg.create_startup_package_node()
        except Exception as e:
            # invert self.add_package
            del self._package_list[codepath]
            raise
        self._dependency_graph.add_vertex(pkg.identifier)
        if pkg.identifier not in self._package_versions:
            self._package_versions[pkg.identifier] = {}
        self._package_versions[pkg.identifier][pkg.version] = pkg
        for old_id in pkg.old_identifiers:
            self._old_identifier_map[old_id] = pkg.identifier
        try:
            self.add_dependencies(pkg)
            #check_requirements is now called in pkg.initialize()
            #pkg.check_requirements()
            self._registry.initialize_package(pkg)
            self._registry.signals.emit_new_package(pkg.identifier, True)
            app.send_notification("package_added", codepath)
            self.add_menu_items(pkg)
            self._startup.set_package_to_enabled(codepath)
        except Exception as e:
            del self._package_versions[pkg.identifier][pkg.version]
            if len(self._package_versions[pkg.identifier]) == 0:
                del self._package_versions[pkg.identifier]
            self.remove_old_identifiers(pkg.identifier)
            self._dependency_graph.delete_vertex(pkg.identifier)
            # invert self.add_package
            del self._package_list[codepath]
            # if we adding the package to the registry, make sure we
            # remove it if initialization fails
            try:
                self._registry.remove_package(pkg)
            except MissingPackage:
                pass
            raise e
        self._startup.save_persisted_startup()

    def late_disable_package(self, codepath):
        """Disables a package 'late', i.e. after VisTrails initialization.

        Note that all the reverse dependencies need to already be disabled.
        """
        pkg = self.get_package_by_codepath(codepath)
        self.remove_package(codepath)
        app = get_vistrails_application()
        self._startup.set_package_to_disabled(codepath)
        self._startup.save_persisted_startup()

    def reload_package_disable(self, codepath):
        # for all reverse dependencies, disable them
        prefix_dictionary = {}
        pkg = self.get_package_by_codepath(codepath)
        reverse_deps = []
        for dep_id in self.all_reverse_dependencies(pkg.identifier):
            reverse_deps.append(self.get_package(dep_id))

        for dep_pkg in reverse_deps:
            prefix_dictionary[dep_pkg.codepath] = dep_pkg.prefix
            self.late_disable_package(dep_pkg.codepath)

        # Queue the re-enabling of the packages so event loop can free
        # any QObjects whose deleteLater() method is invoked
        app = get_vistrails_application()
        app.send_notification("pm_reloading_package", codepath,
                              reverse_deps, prefix_dictionary)
        # self.emit(self.reloading_package_signal,
        #           codepath,
        #           reverse_deps,
        #           prefix_dictionary)

    def reload_package_enable(self, reverse_deps, prefix_dictionary):
        # for all reverse dependencies, enable them
        for dep_pkg in reversed(reverse_deps):
            self.late_enable_package(dep_pkg.codepath, prefix_dictionary)

    def initialize_packages(self, prefix_dictionary={},
                            report_missing_dependencies=True):
        """Initializes all installed packages.

        :param prefix_dictionary: dictionary from package names to the prefix
        such that prefix + package_name is a valid python import.
        :type prefix_dictionary: dict
        """
        failed = []
        # import the modules
        app = get_vistrails_application()
        for package in self._package_list.values():
            # print '+ initializing', package.codepath, id(package)
            if package.initialized():
                # print '- already initialized'
                continue
            try:
                prefix = prefix_dictionary.get(package.codepath)
                if prefix is None:
                    prefix = self._default_prefix_dict.get(package.codepath)
                package.load(prefix)
            except Package.LoadFailed as e:
                debug.critical(
                        "Package %s failed to load and will be disabled" % (
                            package.name or
                            ("<codepath %s>" % package.codepath)),
                        e)
                # We disable the package manually to skip over things
                # we know will not be necessary - the only thing needed is
                # the reference in the package list
                self._startup.set_package_to_disabled(package.codepath)
                failed.append(package)
            except MissingRequirement as e:
                debug.critical("Package <codepath %s> is missing a "
                               "requirement: %s" % (
                                   package.codepath, e.requirement),
                               e)
            except Package.InitializationFailed as e:
                debug.critical("Initialization of package <codepath %s> "
                               "failed and will be disabled" %
                               package.codepath,
                               e)
                # We disable the package manually to skip over things
                # we know will not be necessary - the only thing needed is
                # the reference in the package list
                self._startup.set_package_to_disabled(package.codepath)
                failed.append(package)
            else:
                if package.identifier not in self._package_versions:
                    self._package_versions[package.identifier] = {}
                    self._dependency_graph.add_vertex(package.identifier)
                elif package.version in \
                        self._package_versions[package.identifier]:
                    raise VistrailsInternalError("Duplicate package version: "
                                                 "'%s' (version %s) in %s" % \
                                                     (package.identifier,
                                                      package.version,
                                                      package.codepath))
                else:
                    debug.warning('Duplicate package identifier: %s' % \
                                      package.identifier)
                self._package_versions[package.identifier][package.version] = \
                    package
                for old_id in package.old_identifiers:
                    self._old_identifier_map[old_id] = package.identifier

        for pkg in failed:
            del self._package_list[pkg.codepath]
        failed = []

        # determine dependencies
        for package in self._package_list.values():
            try:
                self.add_dependencies(package)
            except Package.MissingDependency as e:
                if report_missing_dependencies:
                    debug.critical("Dependencies of package %s are missing "
                                   "so it will be disabled" % package.name,
                                   e)
            except Exception as e:
                if report_missing_dependencies:
                    debug.critical("Got an exception while getting dependencies "
                                   "of %s so it will be disabled" % package.name,
                                   e)
            else:
                continue
            self._startup.set_package_to_disabled(package.codepath)
            self._dependency_graph.delete_vertex(package.identifier)
            del self._package_versions[package.identifier][package.version]
            if len(self._package_versions[package.identifier]) == 0:
                del self._package_versions[package.identifier]
            self.remove_old_identifiers(package.identifier)
            failed.append(package)

        for pkg in failed:
            del self._package_list[pkg.codepath]

        # perform actual initialization
        try:
            g = self._dependency_graph.inverse_immutable()
            sorted_packages = g.vertices_topological_sort()
        except vistrails.core.data_structures.graph.GraphContainsCycles as e:
            raise self.DependencyCycle(e.back_edge[0],
                                       e.back_edge[1])

        for name in sorted_packages:
            pkg = self.get_package(name)
            if not pkg.initialized():
                #check_requirements is now called in pkg.initialize()
                #pkg.check_requirements()
                try:
                    self._registry.initialize_package(pkg)
                except MissingRequirement as e:
                    if report_missing_dependencies:
                        debug.critical("Package <codepath %s> is missing a "
                                       "requirement: %s" % (
                                           pkg.codepath, e.requirement),
                                       e)
                    self.late_disable_package(pkg.codepath)
                except Package.InitializationFailed as e:
                    debug.critical("Initialization of package <codepath %s> "
                                   "failed and will be disabled" %
                                   pkg.codepath,
                                   e)
                    # We disable the package manually to skip over things
                    # we know will not be necessary - the only thing needed is
                    # the reference in the package list
                    self.late_disable_package(pkg.codepath)
                else:
                    self.add_menu_items(pkg)
                    app = get_vistrails_application()
                    app.send_notification("package_added", pkg.codepath)

        self._startup.save_persisted_startup()

    def add_menu_items(self, pkg):
        """Emit the appropriate signal if the package has menu items.

        :type pkg: Package
        """
        items = pkg.menu_items()
        if items:
            app = get_vistrails_application()
            app.send_notification("pm_add_package_menu", pkg.identifier,
                                  pkg.name, items)
            # self.emit(self.add_package_menu_signal,
            #           pkg.identifier,
            #           pkg.name,
            #           items)

    def remove_menu_items(self, pkg):
        """Emit the appropriate signal if the package has menu items.

        :type pkg: Package
        """
        if pkg.menu_items():
            app = get_vistrails_application()
            app.send_notification("pm_remove_package_menu",
                                   pkg.identifier)
            # self.emit(self.remove_package_menu_signal,
            #           pkg.identifier)

    def show_error_message(self, pkg, msg):
        """Print and emit a notification for an error message.
        """
        # TODO: isn't debug enough for the UI?
        debug.critical("Package %s (%s) says: %s"%(pkg.name,
                                                   pkg.identifier,
                                                   msg))
        app = get_vistrails_application()
        app.send_notification("pm_package_error_message", pkg.identifier,
                              pkg.name, msg)

    def enabled_package_list(self):
        """Returns list of all enabled packages."""
        return self._package_list.values()

    def identifier_is_available(self, identifier):
        """Searchs for an available (but disabled) package.

        If found, returns succesfully loaded, uninitialized package.

        There can be multiple package versions for a single identifier. If so,
        return the version that passes requirements, or the latest version.
        """
        matches = []
        for codepath in self.available_package_names_list():
            pkg = self.get_available_package(codepath)
            try:
                pkg.load()
                if pkg.identifier == identifier:
                    matches.append(pkg)
                elif identifier in pkg.old_identifiers:
                    matches.append(pkg)
                if (hasattr(pkg._module, "can_handle_identifier") and
                        pkg._module.can_handle_identifier(identifier)):
                    matches.append(pkg)
            except (pkg.LoadFailed, pkg.InitializationFailed,
                    MissingRequirement):
                pass
            except Exception as e:
                debug.warning(
                    "Error loading package <codepath %s>" % pkg.codepath,
                    e)
        if len(matches) == 0:
            return None
        elif len(matches) == 1:
            return matches[0]
        # return package version that passes requirements
        valids = []
        for pkg in matches:
            try:
                pkg.check_requirements()
                valids.append(pkg)
            except (pkg.LoadFailed, pkg.InitializationFailed,
                    MissingRequirement):
                pass
            except Exception as e:
                debug.warning(
                    "Package <codepath %s> raised an exception while "
                    "querying requirements" % pkg.codepath,
                    e)
        if len(valids) == 1:
            return valids[0]
        elif len(valids) == 0:
            # return latest invalid package
            valids = matches
        # return latest version
        return sorted(valids, key=lambda x: LooseVersion(x.version))[-1]

    def available_package_names_list(self):
        """Returns the list of all available packages' codepaths.
        """
        return list(self._available_packages.keys())

    def build_available_package_names_list(self):
        def is_vistrails_package(path):
            if os.path.isfile(path):
                return (path.endswith('.py') and
                        not path.endswith('__init__.py'))
            elif os.path.isdir(path):
                return os.path.isfile(os.path.join(path, '__init__.py'))
            return False

        def search(dirname, prefix):
            for name in os.listdir(dirname):
                if is_vistrails_package(os.path.join(dirname, name)):
                    if name.endswith('.py'):
                        name = name[:-3]
                    self.get_available_package(name, prefix=prefix)

        # Finds standard packages
        packages = self.import_packages_module()
        # This makes VisTrails not zip-safe
        search(os.path.dirname(packages.__file__),
               prefix='vistrails.packages.')

        # Finds user packages
        userpackages = self.import_user_packages_module()
        if userpackages is not None:
            search(os.path.dirname(userpackages.__file__),
                   prefix='userpackages.')

        # Finds plugin packages
        try:
            from pkg_resources import iter_entry_points
        except ImportError:
            pass
        else:
            for entry_point in iter_entry_points('vistrails.packages'):
                # Reads module name and turns it into prefix and codepath
                name = entry_point.module_name.rsplit('.', 1)
                if len(name) > 1:
                    prefix, name = name
                    prefix = '%s.' % prefix
                else:
                    prefix = ''
                    name, = name

                # Create the Package, with the right prefix
                self.get_available_package(name, prefix=prefix)

        return list(self._available_packages.keys())

    def dependency_graph(self):
        """Returns a graph with package dependencies.

        Vertices are the package identifiers (strings), and an edge u -> v
        means that u depends on v.

        :rtype: Graph
        """
        return self._dependency_graph

    def can_be_disabled(self, identifier):
        """Returns whether a package has no reverse dependencies.
        """
        return self._dependency_graph.in_degree(identifier) == 0

    def reverse_dependencies(self, identifier):
        lst = [x[0] for x in
               self._dependency_graph.inverse_adjacency_list[identifier]]
        return lst

    # use this call if we're not necessarily loading
    def build_dependency_graph(self, pkg_identifiers):
        dep_graph = vistrails.core.data_structures.graph.Graph()

        def process_dependencies(identifier):
            dep_graph.add_vertex(identifier)
            pkg = self.identifier_is_available(identifier)
            if pkg:
                deps = pkg.dependencies()
                for dep in deps:
                    if isinstance(dep, tuple):
                        dep_name = dep[0]
                    else:
                        dep_name = dep

                    if dep_name not in self._dependency_graph.vertices and \
                            not dep_graph.has_edge(identifier, dep_name):
                        dep_graph.add_edge(identifier, dep_name)
                        process_dependencies(dep_name)

        for pkg_identifier in pkg_identifiers:
            process_dependencies(pkg_identifier)

        return dep_graph

    def get_ordered_dependencies(self, dep_graph, identifiers=None):
        try:
            sorted_packages = dep_graph.vertices_topological_sort(identifiers)
        except vistrails.core.data_structures.graph.GraphContainsCycles as e:
            raise self.DependencyCycle(e.back_edge[0],
                                       e.back_edge[1])
        return list(reversed(sorted_packages))

    def get_all_dependencies(self, identifier, reverse=False, dep_graph=None):
        if dep_graph is None:
            dep_graph = self._dependency_graph

        if reverse:
            adj_list = dep_graph.inverse_adjacency_list
        else:
            adj_list = dep_graph.adjacency_list

        all = [identifier]
        last_adds = [identifier]
        while len(last_adds) != 0:
            adds = [x[0] for y in last_adds for x in adj_list[y]]
            all.extend(adds)
            last_adds = adds

        seen = set()
        order = []
        for pkg in reversed(all):
            if pkg not in seen:
                order.append(pkg)
                seen.add(pkg)
        return order

    def all_dependencies(self, identifier, dep_graph=None):
        return self.get_all_dependencies(identifier, False, dep_graph)

    def all_reverse_dependencies(self, identifier, dep_graph=None):
        return self.get_all_dependencies(identifier, True, dep_graph)

def get_package_manager():
    if not _package_manager:
        raise VistrailsInternalError("package manager not constructed yet.")
    return _package_manager

##############################################################################

import unittest


class TestImports(unittest.TestCase):
    def test_package(self):
        from vistrails.tests.utils import MockLogHandler

        pm = get_package_manager()
        pm.get_available_package(
                'test_import_pkg',
                prefix='vistrails.tests.resources.import_pkg.')

        # Check the package is in the list
        available_pkg_names = pm.available_package_names_list()
        self.assertIn('test_import_pkg', available_pkg_names)

        # Import __init__ and check metadata
        pkg = pm.look_at_available_package('test_import_pkg')
        with MockLogHandler(debug.DebugPrint.getInstance().logger) as log:
            pkg.load('vistrails.tests.resources.import_pkg.')
        self.assertEqual(len(log.messages['warning']), 1)
        self.assertEqual(pkg.identifier,
                         'org.vistrails.tests.test_import_pkg')
        self.assertEqual(pkg.version,
                         '0.42')
        for n in ['vistrails.tests.resources.import_targets.test1',
                  'vistrails.tests.resources.import_targets.test2']:
            self.assertIn(n, sys.modules, "%s not in sys.modules" % n)

        # Import init.py
        pm.late_enable_package(
                'test_import_pkg',
                {'test_import_pkg':
                 'vistrails.tests.resources.import_pkg.'})
        pkg = pm.get_package_by_codepath('test_import_pkg')
        for n in ['vistrails.tests.resources.import_targets.test3',
                  'vistrails.tests.resources.import_targets.test4',
                  'vistrails.tests.resources.import_targets.test5']:
            self.assertIn(n, sys.modules, "%s not in sys.modules" % n)

        # Check dependencies
        deps = pkg.get_py_deps()
        for dep in ['vistrails.tests.resources.import_pkg.test_import_pkg',
                    'vistrails.tests.resources.import_pkg.test_import_pkg.init',
                    'vistrails.tests.resources.import_pkg.test_import_pkg.module1',
                    'vistrails.tests.resources.import_pkg.test_import_pkg.module2',
                    'vistrails.tests.resources.import_targets',
                    'vistrails.tests.resources.import_targets.test1',
                    'vistrails.tests.resources.import_targets.test2',
                    'vistrails.tests.resources.import_targets.test3',
                    'vistrails.tests.resources.import_targets.test4',
                    'vistrails.tests.resources.import_targets.test5',
                    'vistrails.tests.resources.import_targets.test6']:
            self.assertIn(dep, deps)



###############################################################################
##
## Copyright (C) 2014-2016, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" The file describes a container widget consisting of a pipeline
view and a version tree for each opened Vistrail """
from PyQt5 import QtCore, QtWidgets


from vistrails.core import debug
from vistrails.core.collection import Collection
from vistrails.core.system import vistrails_default_file_type, \
    vistrails_file_directory
from vistrails.core.thumbnails import ThumbnailCache
from vistrails.core.vistrail.vistrail import Vistrail
from vistrails.core.vistrail.pipeline import Pipeline
from vistrails.core.log.log import Log
from vistrails.core.log.opm_graph import OpmGraph
from vistrails.core.log.prov_document import ProvDocument
from vistrails.core.db.locator import XMLFileLocator
from vistrails.core.modules.module_registry import ModuleRegistry
from vistrails.core.configuration import get_vistrails_configuration

from vistrails.gui.collection.vis_log import QLogView
from vistrails.gui.common_widgets import QMouseTabBar
from vistrails.gui.mashups.mashup_view import QMashupView
from vistrails.gui.module_info import QModuleInfo
from vistrails.gui.paramexplore.pe_view import QParamExploreView
from vistrails.gui.pipeline_view import QPipelineView
from vistrails.gui.ports_pane import ParameterEntry
from vistrails.gui.query_view import QQueryView, QueryEntry,\
    QQueryResultWorkflowView
from vistrails.gui.version_view import QVersionTreeView
from vistrails.gui.vis_diff import QDiffView
from vistrails.gui.vistrail_controller import VistrailController

################################################################################

class QVistrailView(QtWidgets.QWidget):
    """
    QVistrailView is a widget containing four stacked widgets: Pipeline View,
    Version Tree View, Query View and Parameter Exploration view
    for manipulating vistrails.
    """
    def __init__(self, vistrail, locator=None, abstraction_files=None,
                 thumbnail_files=None, mashups=None, parent=None):
        """ QVistrailView(parent: QWidget) -> QVistrailView
        
        """
        super().__init__(parent=parent)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.is_executing = False
        self.is_abstraction = False
        self.notifications = {}
        self.tabs = QMouseTabBar(self)
        self.tabs.setDocumentMode(True)
        self.tabs.setTabsClosable(True)
        self.tabs.setExpanding(False)
        #self.tabs.setMovable(True)
        self.tabs.hide()
        self.tabs.tabDoubleClicked.connect(self.tabDoubleClicked)
        layout.addWidget(self.tabs)
        self.stack = QtWidgets.QStackedWidget(self)
        layout.addWidget(self.stack)
        self.setLayout(layout)

        #this index is for pipeline/diff views
        self.tab_to_stack_idx = {}
        self.tab_state = {}
        self.tab_to_view = {}
        #self.button_to_tab_idx = Bidict()
        self.detached_views = {}

        # Create the initial views
        self.version_view = None
        pipeline_view = self.create_pipeline_view(set_controller=False)
        self.version_view = self.create_version_view()
        self.query_view = self.create_query_view()
        self.pe_view = self.create_pe_view()
        self.log_view = self.create_log_view()
        self.mashup_view = self.create_mashup_view(False)

        # Initialize the vistrail controller
        self.locator = locator
        self.controller = VistrailController(vistrail, 
                                             self.locator, 
                                             abstractions=abstraction_files, 
                                             thumbnails=thumbnail_files,
                                             mashups=mashups,
                                             pipeline_view=pipeline_view)
        
        self.set_controller(self.controller)
        pipeline_view.set_to_current()

        self.tabs.setCurrentIndex(0)
        self.current_tab = self.stack.setCurrentIndex(0)
        self.pipeline_selected()
        self.tabs.currentChanged.connect(self.tab_changed)
        self.tabs.tabCloseRequested.connect(self.tab_close_request)
       
        #self.view_changed()
        #self.tab_changed(0)

        self.controller.stateChanged.connect(self.stateChanged)

        from vistrails.gui.vistrails_window import _app
        _app.register_notification("reg_new_abstraction", 
                                   self.controller.check_subworkflow_versions)
        _app.register_notification("reg_deleted_abstraction",
                                   self.controller.check_subworkflow_versions)

        # self.controller = VistrailController()
        # self.controller.vistrail_view = self
        # self.connect(self.controller,
        #              QtCore.SIGNAL('stateChanged'),
        #              self.stateChanged)
        # self.connect(self.controller,
        #              QtCore.SIGNAL('new_action'),
        #              self.new_action)

        # # self.versionTab.versionView.scene()._vistrail_view = self
        # self.connect(self.versionTab.versionView.scene(),
        #              QtCore.SIGNAL('versionSelected(int,bool,bool,bool)'),
        #              self.versionSelected,
        #              QtCore.Qt.QueuedConnection)

        # self.connect(self.versionTab,
        #              QtCore.SIGNAL('twoVersionsSelected(int,int)'),
        #              self.twoVersionsSelected)
        # self.connect(self.queryTab,
        #              QtCore.SIGNAL('queryPipelineChange'),
        #              self.queryPipelineChange)
        # self.connect(self.peTab,
        #              QtCore.SIGNAL('exploreChange(bool)'),
        #              self.exploreChange)

        # # We also keep track where this vistrail comes from
        # # So we can save in the right place
        # self.locator = None
        
        # self.closeEventHandler = None

        # # Keep the state of the execution button and menu items for the view
        # self.execQueryEnabled = False
        # self.execDiffEnabled = False
        # self.execExploreEnabled = False
        # self.execPipelineEnabled = False
        # self.execDiffId1 = -1
        # self.execDiffId2 = -1
        if get_vistrails_configuration().detachHistoryView:
            self.detach_history_view()

    def get_notifications(self):
        return self.notifications

    def set_notification(self, notification_id, method):
        if notification_id not in self.notifications:
            self.notifications[notification_id] = []
        self.notifications[notification_id].append(method)

    def set_controller(self, controller):
        self.controller = controller
        self.controller.vistrail_view = self
        for i in range(self.stack.count()):
            view = self.stack.widget(i)
            if hasattr(view, 'set_controller'):
                view.set_controller(controller)

    def get_controller(self):
        return self.controller

    def get_name(self):
        title = self.controller.name
        if title=='':
            title = 'Untitled%s'%vistrails_default_file_type()
        if self.controller.changed:
            title += '*'
        # self.setWindowTitle(title)
        return title

    def set_name(self):
        title = self.get_name()
        self.setWindowTitle(title)

    def reset_version_view(self):
        from vistrails.gui.vistrails_window import _app
        if self.version_view is not None:
            select_node = True
            if _app._previous_view and _app._previous_view in self.detached_views:
                select_node = False
            self.version_view.scene().setupScene(self.controller, select_node)

    def reset_tab_state(self):
        try:
            qaction = self.tab_state[self.tabs.currentIndex()]
            qaction.trigger()
        except Exception as e:
            debug.unexpected_exception(e)
        
    def reset_tab_view_to_current(self):
        index = self.tabs.currentIndex()
        view = self.stack.widget(self.tab_to_stack_idx[index])
        #print "view changed: ", view
        self.set_to_current(view)

    def showEvent(self, event):
        self.stack.currentWidget().viewSelected()

    def pipeline_selected(self):
        from vistrails.gui.vistrails_window import _app
        if hasattr(self.window(), 'qactions'):
            window = self.window()
        else:
            window = _app
        #print "PIPELINE"
        self.stack.setCurrentIndex(
            self.tab_to_stack_idx[self.tabs.currentIndex()])
        self.tabs.setTabText(self.tabs.currentIndex(),
                             self.stack.currentWidget().get_title())
        self.tab_state[self.tabs.currentIndex()] = window.qactions['pipeline']
        self.tab_to_view[self.tabs.currentIndex()] = self.get_current_tab()
        self.stack.currentWidget().viewSelected()

    def pipeline_unselected(self):
        #print "PIPELINE UN"
        self.stack.setCurrentIndex(
            self.tab_to_stack_idx[self.tabs.currentIndex()])
        self.tabs.setTabText(self.tabs.currentIndex(),
                             self.stack.currentWidget().get_title())

    def history_selected(self):
        from vistrails.gui.vistrails_window import _app
        if get_vistrails_configuration().detachHistoryView:
            _app.history_view.raise_()
            return
        if hasattr(self.window(), 'qactions'):
            window = self.window()
        else:
            window = _app
        #print "VERSION"
        self.stack.setCurrentIndex(self.stack.indexOf(self.version_view))
        self.tabs.setTabText(self.tabs.currentIndex(), "History")
        self.tab_state[self.tabs.currentIndex()] = window.qactions['history']
        self.tab_to_view[self.tabs.currentIndex()] = self.get_current_tab()
        self.stack.currentWidget().viewSelected()

    def history_unselected(self):
        #print "VERSION UN"
        self.stack.setCurrentIndex(
            self.tab_to_stack_idx[self.tabs.currentIndex()])
        self.tabs.setTabText(self.tabs.currentIndex(), 
                             self.stack.currentWidget().get_title())

    def query_selected(self):
        from vistrails.gui.vistrails_window import _app
        if hasattr(self.window(), 'qactions'):
            window = self.window()
        else:
            window = _app
        #print "QUERY"
        self.stack.setCurrentIndex(self.stack.indexOf(self.query_view))
        self.tabs.setTabText(self.tabs.currentIndex(), "Search")
        self.tab_state[self.tabs.currentIndex()] = window.qactions['search']
        self.tab_to_view[self.tabs.currentIndex()] = self.get_current_tab()

    def query_unselected(self):
        #print "QUERY UN"
        self.stack.setCurrentIndex(
            self.tab_to_stack_idx[self.tabs.currentIndex()])
        self.tabs.setTabText(self.tabs.currentIndex(), 
                             self.stack.currentWidget().get_title())

    def explore_selected(self):
        from vistrails.gui.vistrails_window import _app
        if hasattr(self.window(), 'qactions'):
            window = self.window()
        else:
            window = _app
        #print "EXPLORE"
        self.stack.setCurrentIndex(self.stack.indexOf(self.pe_view))
        self.tabs.setTabText(self.tabs.currentIndex(),
                             self.stack.currentWidget().get_title())
        self.tab_state[self.tabs.currentIndex()] = window.qactions['explore']
        self.tab_to_view[self.tabs.currentIndex()] = self.get_current_tab()

    def explore_unselected(self):
        #print "EXPLORE UN"
        self.stack.setCurrentIndex(
            self.tab_to_stack_idx[self.tabs.currentIndex()])
        self.tabs.setTabText(self.tabs.currentIndex(), 
                             self.stack.currentWidget().get_title())

    def provenance_selected(self):
        from vistrails.gui.vistrails_window import _app
        if hasattr(self.window(), 'qactions'):
            window = self.window()
        else:
            window = _app
        #print "PROVENANCE"
        self.stack.setCurrentIndex(self.stack.indexOf(self.log_view))
        self.tabs.setTabText(self.tabs.currentIndex(), "Provenance")
        self.tab_state[self.tabs.currentIndex()] = window.qactions['provenance']
        self.tab_to_view[self.tabs.currentIndex()] = self.get_current_tab()

    def provenance_unselected(self):
        #print "PROVENANCE UN"
        self.stack.setCurrentIndex(
            self.tab_to_stack_idx[self.tabs.currentIndex()])
        self.tabs.setTabText(self.tabs.currentIndex(), 
                             self.stack.currentWidget().get_title())

    def mashup_selected(self):
        from vistrails.gui.vistrails_window import _app
        if hasattr(self.window(), 'qactions'):
            window = self.window()
        else:
            window = _app
        #print "MASHUP"
        #print self.stack.count(), self.stack.indexOf(self.mashup_view)
        try:
            self.stack.setCurrentIndex(self.stack.indexOf(self.mashup_view))
            self.tabs.setTabText(self.tabs.currentIndex(), "Mashup")
            self.tab_state[self.tabs.currentIndex()] = window.qactions['mashup']
            self.mashup_view.updateView()
            self.tab_to_view[self.tabs.currentIndex()] = self.get_current_tab()
        except Exception as e:
            debug.unexpected_exception(e)
            print("EXCEPTION: ", debug.format_exception(e))
    def mashup_unselected(self):
        #print "MASHUP UN"
        self.stack.setCurrentIndex(
            self.tab_to_stack_idx[self.tabs.currentIndex()])
        self.tabs.setTabText(self.tabs.currentIndex(), 
                             self.stack.currentWidget().get_title())
        
    def pipeline_change(self, checked):
        if checked:
            #print "PIPELINE SELECTED"
            self.pipeline_selected()
        else:
            #print "PIPELINE UNSELECTED"
            self.pipeline_unselected()
        self.view_changed()

    def history_change(self, checked):
        from .vistrails_window import _app
        if checked:
            #print "HISTORY SELECTED"
            self.history_selected()
        else:
            #print "HISTORY UNSELECTED"
            self.history_unselected()
        self.view_changed()

    def search_change(self, checked):
        if checked:
            self.query_selected()
        else:
            self.query_unselected()
        self.view_changed()

    def explore_change(self, checked):
        if checked:
            self.explore_selected()
        else:
            self.explore_unselected()
        self.view_changed()

    def provenance_change(self, checked):
        if checked:
            self.provenance_selected()
        else:
            self.provenance_unselected()
        self.view_changed()

    def mashup_change(self, checked):
        if checked:
            self.mashup_selected()
        else:
            self.mashup_unselected()
        self.view_changed()
    
    def show_group(self):
        pipelineScene = self.controller.current_pipeline_scene
        items = pipelineScene.get_selected_item_ids(True)
        if items is not None:
            for m_id in items[0]:
                module = pipelineScene.current_pipeline.modules[m_id]
                if module.is_group() or module.is_abstraction():
                    newPipelineView = self.add_pipeline_view(True)
                    newPipelineView.controller.current_pipeline_view = \
                        newPipelineView
                    module.pipeline.ensure_connection_specs()
                    newPipelineView.scene().setupScene(module.pipeline)

    def create_view(self, klass, add_tab=True):
        view = klass(self)
        view.set_vistrail_view(self)
        idx = self.stack.addWidget(view)
        view.set_index(idx)
        if add_tab:
            tab_idx = self.tabs.addTab(view.get_title())
            view.set_tab_idx(tab_idx)
            self.tab_to_stack_idx[tab_idx] = idx
            self.tab_to_view[tab_idx] = view
            if self.isTabDetachable(tab_idx):
                self.tabs.setTabToolTip(tab_idx, "Double-click to detach it")
            
        view.windowTitleWasChanged.connect(self.view_title_changed)
        if self.tabs.count() == 1:
            #self.tabs.hide()
            self.tabs.setTabsClosable(False)
        else:
            self.tabs.setTabsClosable(True)
        self.updateTabsTooTip()
        self.tabs.show()
        return view

    def detach_history_view(self):
        from vistrails.gui.vistrails_window import _app
        view = self.version_view
        window = _app.history_view
        self.version_index = window.stack.addWidget(view)
        window.stack.setCurrentIndex(self.version_index)
        window.view = view

    def detach_view(self, tab_idx):
        from vistrails.gui.vistrails_window import QBaseViewWindow
        if tab_idx in self.tab_to_stack_idx:
            stack_index = self.tab_to_stack_idx[tab_idx]
            view = self.stack.widget(stack_index)
            title = view.get_long_title()
            self.remove_view_by_index(tab_idx)
            window = QBaseViewWindow(view=view, parent=None)
            view.set_title(title)
            window.setWindowTitle(title)
            window.viewWasClosed.connect(self.detachedViewWasClosed)
            self.detached_views[view] = window
            window.move(self.rect().center())
            window.show()
        else:
            print("Error detach_view: ", tab_idx, self.tab_to_stack_idx)
    
    def isTabDetachable(self, index):
        if index in self.tab_to_view:
            return self.tabs.count() > 1 and self.tab_to_view[index].detachable
        return False
    
    def closeDetachedViews(self):
        windows = list(self.detached_views.values())
        for w in windows:
            if w:
                w.close()
            
    def detachedViewWasClosed(self, view):
        if self.controller.current_pipeline_view == view:
            self.controller.current_pipeline_view = None
            self.activateWindow()
            self.reset_tab_view_to_current()
            self.view_changed()
        del self.detached_views[view]
        
    def updateTabsTooTip(self):
        for i in range(self.tabs.count()):
            if self.isTabDetachable(i):
                self.tabs.setTabToolTip(i, "Double-click to detach it")
            else:
                self.tabs.setTabToolTip(i, "")
    
    def tabDoubleClicked(self, index, pos):
        if self.isTabDetachable(index):
            self.detach_view(index)
            
    def view_title_changed(self, view):
        if self.stack.currentWidget() == view:
            self.tabs.setTabText(self.tabs.currentIndex(), view.windowTitle())

    def update_indexes(self, rm_tab_idx, rm_stack_idx):
        for (t,s) in self.tab_to_stack_idx.items():
            if s > rm_stack_idx:
                self.tab_to_stack_idx[t] -= 1
        tabs = list(self.tab_to_stack_idx.keys())
        tabs.sort()
        for t in tabs:
            if t > rm_tab_idx:
                self.tab_to_stack_idx[t-1] = self.tab_to_stack_idx[t]
                self.tab_state[t-1] = self.tab_state[t]
        del self.tab_to_stack_idx[tabs[-1]]
        del self.tab_state[tabs[-1]]
        for idx in range(self.stack.count()):
            if idx >= rm_stack_idx:
                view = self.get_tab(idx)
                view.set_index(idx)
                if view.tab_idx > rm_tab_idx:
                    view.set_tab_idx(view.tab_idx-1)
    
    def tab_close_request(self, index):
        """ Ignore if this is the last pipeline tab

            If there is one pipeline tab and one diff tab, close buttons are
            visible for both of them, but we should not allow the closing of
            this last pipeline tab.
        """
        def is_pipeline_tab(view):
            # QDiffView is currently the only non-pipeline tab
            return not isinstance(view, QDiffView)
        pipeline_tabs = sum([(1 if is_pipeline_tab(t) else 0)
                            for t in self.tab_to_view.values()])
        if pipeline_tabs > 1 or not is_pipeline_tab(self.tab_to_view[index]):
            self.remove_view_by_index(index)

    def remove_view_by_index(self, index):
        self.tabs.currentChanged.disconnect(self.tab_changed)
        close_current = False
        if index == self.tabs.currentIndex():
            close_current = True
        stack_idx = self.tab_to_stack_idx[index]
        #print "\n\n >>>>> remove_view_by_index ", index, stack_idx, self.tabs.currentIndex()
        self.tabs.removeTab(index)
        del self.tab_to_view[index]
        if stack_idx >= 0:
            view = self.stack.widget(stack_idx)
            view.windowTitleWasChanged.disconnect(self.view_title_changed)
            self.stack.removeWidget(view)
        self.update_indexes(index, stack_idx)
        if self.tabs.count() == 1:
            self.tabs.setTabsClosable(False)
            self.updateTabsTooTip()
        if close_current:
            if index >= self.tabs.count():
                new_index = index - 1
            else:
                new_index = index
        
            self.tab_changed(new_index)
        self.tabs.currentChanged.connect(self.tab_changed)
#        self.tabs.setCurrentIndex(new_index)
#        print self.current_tab
        
#        self.view_changed()
        

    def switch_to_tab(self, index):
#        if index < 0:
#            index = self.tabs.count() + index
        self.tabs.setCurrentIndex(index)
        self.tab_changed(index)

    def get_current_tab(self, query_top_level=False):
        window = QtWidgets.QApplication.activeWindow()
        if window in list(self.detached_views.values()):
            return window.view   
        else:
            #if none of the detached views is active we will assume that the
            #window containing this vistrail has focus
            widget = self.stack.currentWidget()
            if not query_top_level and isinstance(widget, QQueryView):
                widget = widget.get_current_view()
            return widget
        
    def get_current_outer_tab(self):
        window = QtWidgets.QApplication.activeWindow()
        if window in list(self.detached_views.values()):
            return window.view   
        else:
            #if none of the detached views is active we will assume that the
            #window containing this vistrail has focus
            return self.stack.currentWidget()
        
    def get_tab(self, stack_idx):
        widget = self.stack.widget(stack_idx)
        if isinstance(widget, QQueryView):
            widget = widget.get_current_view()
        return widget

    def view_changed(self):
        from vistrails.gui.vistrails_window import _app
        #view = self.stack.currentWidget()
        view = self.get_current_outer_tab()
        #print "changing tab from: ",self.current_tab, " to ", view
        #print self.tab_to_stack_idx
        if view != self.current_tab:
            _app.closeNotPinPalettes()
            #print "!!unset_action_links of ", self.current_tab
            _app.unset_action_links(self.current_tab)
            self.current_tab = view
#            print "\n!! _app.notifications: "
#            for (k, v) in _app.notifications.iteritems():
#                print "   ", k, "  (%s) "%len(v)
#                for m in v: 
#                    print "     ", m
#            print "\n!!set_action_defaults of ", self.current_tab
            _app.set_action_defaults(self.current_tab)
            #print "\n!!set_action_links of ", self.current_tab 
            _app.set_action_links(self.current_tab.action_links, self.current_tab,
                                  self)
            self.showCurrentViewPalettes()

        #else:
           # print "tabs the same. do nothing"
        if isinstance(view, QQueryView):
            _app.notify("controller_changed", view.p_controller)
            if view.current_display != QQueryView.VISUAL_SEARCH_VIEW:
                # Force result version prop
                view.result_version_selected(view.vt_controller.current_version, False)
            _app.notify("entry_klass_changed", QueryEntry)
        else:
            _app.notify("entry_klass_changed", ParameterEntry)
            _app.notify("controller_changed", self.controller)

        if self.window().isActiveWindow():
            if self.isTabDetachable(self.tabs.currentIndex()):
                self.tabs.setTabToolTip(self.tabs.currentIndex(),
                                        "Double-click to detach it")
            else:
                self.tabs.setTabToolTip(self.tabs.currentIndex(),
                                        "")
        if get_vistrails_configuration().detachHistoryView:
            if hasattr(self.window(), 'qactions'):
                _app = self.window()
            _app.history_view.stack.setCurrentIndex(self.version_index)

    def showCurrentViewPalettes(self):
        current_tab = self.get_current_tab(True)
        for dock_loc, palette_klass in current_tab.palette_layout.items():
            palette_instance = palette_klass.instance()
            window = palette_instance.toolWindow().parentWidget()
            if window:
                current_loc = window.dockWidgetArea(palette_instance.toolWindow())
            else:
                current_loc = QtCore.Qt.NoDockWidgetArea
            #print ">> P:", palette_instance.__class__.__name__, current_loc, \
            #        dock_loc
            
            if current_loc == dock_loc:
                # palette_instance.get_action().trigger()
                palette_instance.set_visible(True)
                    
    def tab_changed(self, index):
        #print 'raw tab_changed', index
        if index < 0 or self.controller is None:
            return

        from vistrails.gui.vistrails_window import _app, QVistrailViewWindow

        self.stack.setCurrentIndex(self.tab_to_stack_idx[index])
        if isinstance(self.window(),QVistrailViewWindow):
            window = self.window()
        else:
            window = _app
        #print window
        for action in window.view_action_group.actions():
            action.setChecked(False)
        self.selected_mode = None
        action = None
        if index in self.tab_state:
            action = self.tab_state[index]
            # if action is not None:
                # print 'running toggle'
                # action.toggle()
                # action.setChecked(True)
        else:
            self.tab_state[index] = window.qactions['pipeline']
        if action is not None:
            action.setChecked(True)
            # _app.view_triggered(action)

        view = self.stack.widget(self.tab_to_stack_idx[index])
        #print "view changed: ", view
        self.set_to_current(view)
        
    def set_to_current(self, view):
        from vistrails.gui.vistrails_window import _app
        if isinstance(view, QDiffView):
            view.set_to_current()
            #print "view changed!", self.controller, \
            #    self.controller.current_version
            _app.notify("controller_changed", self.controller)
            self.reset_version_view()
        elif isinstance(view, QLogView):
            view.set_to_current()
            #print "view changed!", self.controller, \
            #    self.controller.current_version
            _app.notify("controller_changed", self.controller)
            self.reset_version_view()
        elif isinstance(view, QPipelineView):
            #print "PIPELINE_VIEW NEW SCENE:", id(view.scene())

            # need to set the controller's version, pipeline, view
            # to this view...
            # self.controller.current_version = view.current_version
            # self.controller.current_pipeline = view.current_pipeline
            view.set_to_current()
            #print "view changed!", self.controller, \
            #    self.controller.current_version
            real_view = self.stack.currentWidget()
            if isinstance(real_view, QQueryView):
                _app.notify("controller_changed", real_view.p_controller)
            else:
                _app.notify("controller_changed", self.controller)
            self.reset_version_view()

    def create_pipeline_view(self, read_only=False, set_controller=True):
        view = self.create_view(QPipelineView)
        view.setReadOnlyMode(read_only)
        view.scene().moduleSelected.connect(self.gen_module_selected(view))
        if set_controller:
            view.set_controller(self.controller)
            view.set_to_current()
        self.set_notification('module_done_configure', view.done_configure)
        #self.switch_to_tab(view.tab_idx)
        return view

    def add_pipeline_view(self, read_only=False):
        view = self.create_pipeline_view(read_only)
        self.switch_to_tab(view.tab_idx)
        return view

    def create_version_view(self):
        view = self.create_view(QVersionTreeView, False)
        view.scene().versionSelected.connect(self.version_selected)
        view.scene().diffRequested.connect(self.diff_requested)
        return view

    def create_query_view(self):
        view = self.create_view(QQueryView, False)
        view.pipeline_view.scene().moduleSelected.connect(
                                 self.gen_module_selected(view.pipeline_view))
        # FIXME: How to make module info read-only
        view.workflow_result_view.scene().moduleSelected.connect(
                          self.gen_module_selected(view.workflow_result_view))
        # self.connect(view.version_result_view.scene(),
        #              QtCore.SIGNAL('versionSelected(int,bool,bool,bool,bool)'),
        #              self.version_selected)
        # self.connect(view.version_result_view.scene(),
        #              QtCore.SIGNAL('diffRequested(int,int)'),
        #              self.diff_requested)
        self.set_notification('query_changed', view.query_changed)
        # DAK: removed this notification: query view has own version
        # self.set_notification('version_changed', view.version_changed)
        return view

    def create_diff_view(self):
        view = self.create_view(QDiffView)
        view.scene().moduleSelected.connect(self.gen_module_selected(view))
        return view

    def create_pe_view(self):
        view = self.create_view(QParamExploreView, False)
        self.set_notification('controller_changed', view.set_controller)
        self.set_notification('pipeline_changed', view.updatePipeline)
        self.set_notification('exploration_changed', view.set_exploration)
        return view

    def create_log_view(self):
        from vistrails.gui.vistrails_window import _app
        view = self.create_view(QLogView, False)
        self.set_notification('execution_changed', view.execution_changed)
        return view
    
    def create_mashup_view(self, set_controller=True):
        #print "******* create mashup view"
        from vistrails.gui.vistrails_window import _app
        view = self.create_view(QMashupView, False)
        if set_controller:
            view.set_controller(self.controller)
        self.set_notification('controller_changed', view.controllerChanged)
        self.set_notification('alias_changed', view.aliasChanged)
        self.set_notification('version_changed', view.versionChanged)
        return view
    
    def gen_module_selected(self, view):
        def module_selected(module_id, selection = []):
            from vistrails.gui.version_prop import QVersionProp
            from vistrails.gui.vistrails_window import _app
            pipeline = view.scene().current_pipeline
            if pipeline is not None and module_id in pipeline.modules:
                module = pipeline.modules[module_id]
                _app.notify('module_changed', module)
                # show module info if both are tabified and
                # workflow info is visible
                if not QModuleInfo.instance().toolWindow().isFloating() and \
                   not QVersionProp.instance().toolWindow().isFloating() and \
                   not QVersionProp.instance().toolWindow().visibleRegion().isEmpty():
                    QModuleInfo.instance().set_visible(True)
            else:
                if isinstance(view, QQueryResultWorkflowView):
                    # we need to set version prop directly
                    QVersionProp.instance().updateController(view.controller)
                    QVersionProp.instance().updateVersion(view.controller.current_version)
                else:
                    _app.notify('module_changed', None)
                # show workflow info if both are tabified and
                # module info is visible
                if not QModuleInfo.instance().toolWindow().isFloating() and \
                   not QVersionProp.instance().toolWindow().isFloating() and \
                   not QModuleInfo.instance().toolWindow().visibleRegion().isEmpty():
                    QVersionProp.instance().set_visible(True)
        return module_selected

    def version_selected(self, version_id, by_click, do_validate=True,
                         from_root=False, double_click=False):
        from vistrails.gui.vistrails_window import _app
        from vistrails.gui.vis_diff import QDiffView
        if hasattr(self.window(), 'qactions'):
            window = self.window()
        else:
            window = _app
        #print 'got version selected:', version_id
        if _app._focus_owner in list(self.detached_views.values()):
            view = _app._focus_owner.view
        elif _app._previous_view in self.detached_views:
            view = _app._previous_view
        else:
            view = self.stack.widget(
                self.tab_to_stack_idx[self.tabs.currentIndex()])
        if view and by_click:
            self.controller.change_selected_version(version_id, True, 
                                                    do_validate, from_root)

            view.scene().fitToView(view, True)
            if double_click:
                # view = self.create_pipeline_view()
                # view.set_controller(self.controller)
                # view.set_to_current()
                # self.tabs.setCurrentWidget(view.parent())
                window.qactions['pipeline'].trigger()
            self.controller.reset_redo_stack()
        if view and not isinstance(view, QDiffView):
            if view not in self.detached_views:
                view.set_title("Pipeline: %s" %
                               self.controller.get_pipeline_name())
            else:
                view.set_title(view.get_long_title())
                view.window().setWindowTitle(view.get_long_title())
        _app.notify("version_changed", self.controller.current_version)
        _app.notify("pipeline_changed", self.controller.current_pipeline)

    def query_version_selected(self, search=None, version_id=None):
        search.cur_controller = self.controller
        if version_id is None:
            self.query_view.set_result_level(
                self.query_view.query_controller.LEVEL_VISTRAIL)
            self.query_view.query_controller.set_search(search)
        else:
            self.query_view.set_result_level(
                self.query_view.query_controller.LEVEL_WORKFLOW)
            self.query_view.query_controller.set_search(search)
            self.query_view.result_version_selected(version_id, True, 
                                                    double_click=True)

        window = self.window()
        window.qactions['search'].trigger()
        
    def diff_requested(self, version_a, version_b, controller_b=None):
        """diff_requested(self, id, id, Vistrail) -> None
        
        Request a diff between two versions.  If controller_b is
        specified, the second version will be derived from that
        vistrail instead of the common vistrail controlled by this
        view.
        """
        controller_b = controller_b or self.controller
        # Upgrade both versions if hiding upgrades
        if getattr(get_vistrails_configuration(), 'hideUpgrades', True):
            version_a = self.controller.create_upgrade(version_a, delay_update=True)
            version_b = controller_b.create_upgrade(version_b, delay_update=True)

        view = self.create_diff_view()
        view.set_controller(self.controller)
        view.set_diff(version_a, version_b, controller_b.vistrail)
        self.switch_to_tab(view.tab_idx)
        view.scene().fitToView(view, True)
        self.controller.check_delayed_update()
        controller_b.check_delayed_update()
        self.view_changed()

    def save_vistrail(self, locator_class, force_choose_locator=False, export=False):
        """
        force_choose_locator=True triggers 'save as' behavior
        export=True does not update the current controller

        """
        locator = self.controller.locator
        if locator_class is None:
            if locator is not None:
                locator_class = type(locator)
            else:
                debug.critical('Failed to save vistrail, unknown '
                               'locator class.')

        #print "CALLED SAVE VISTRAIL", locator_class

        self.flush_changes()
        self.controller.flush_delayed_actions()
        gui_get = locator_class.save_from_gui
        # get a locator to write to
        if not locator or locator.is_untitled():
            force_choose_locator = True
        if force_choose_locator:
            locator = gui_get(self, Vistrail.vtType,
                              self.controller.locator)
        # if couldn't get one, ignore the request
        if not locator:
            return False
        try:
            self.controller.write_vistrail(locator, export=export)
        except Exception:
            debug.critical('Failed to save vistrail', debug.format_exc())
            raise
        if export:
            return self.controller.locator
        
        if not force_choose_locator:
            from vistrails.gui.vistrails_window import _app
            _app.view_changed(self)
            _app.notify("vistrail_saved")
            return locator
        # update collection
        try:
            thumb_cache = ThumbnailCache.getInstance()
            self.controller.vistrail.thumbnails = \
                self.controller.find_thumbnails(
                    tags_only=thumb_cache.conf.tagsOnly)
            self.controller.vistrail.abstractions = \
                self.controller.find_abstractions(self.controller.vistrail, 
                                                  True)
            self.controller.vistrail.mashups = self.controller._mashups

            collection = Collection.getInstance()
            url = locator.to_url()
            entity = collection.updateVistrail(url, self.controller.vistrail)
            # add to relevant workspace categories
            collection.add_to_workspace(entity)
            collection.commit()
        except Exception:
            debug.critical('Failed to index vistrail', debug.format_exc())

        from vistrails.gui.vistrails_window import _app
        # update recent files menu items
        if not self.is_abstraction:
            _app.set_current_locator(locator)
        _app.view_changed(self)
        _app.notify("vistrail_saved")
        # reload workspace entry
        from vistrails.gui.collection.workspace import QWorkspaceWindow
        QWorkspaceWindow.instance().add_vt_window(self)
        return locator

    def save_vistrail_as(self, locator_class):
        #print "CALLED SAVE AS VISTRAIL", locator_class
        self.save_vistrail(locator_class, force_choose_locator=True)

    def export_vistrail(self, locator_class):
        """ Exports vistrail without updating the current vistrail """
        self.save_vistrail(locator_class, force_choose_locator=True, export=True)

    def export_stable(self, locator_class=XMLFileLocator):
        """ save workflow to previous stable version """
        self.flush_changes()
        gui_get = locator_class.save_from_gui
        locator = gui_get(self, Pipeline.vtType)
        if not locator:
            return False
        self.controller.write_workflow(locator, '1.0.3')
        return True

    # FIXME normalize workflow/log/registry!!!
    def save_workflow(self, locator_class, force_choose_locator=True):
        self.flush_changes()
        gui_get = locator_class.save_from_gui
        if force_choose_locator:
            locator = gui_get(self, Pipeline.vtType, self.controller.locator)
        else:
            locator = (self.controller.locator or
                       gui_get(self, Pipeline.vtType,
                               self.controller.locator))
        if locator is not None and locator.is_untitled():
            locator = gui_get(self, Pipeline.vtType, self.controller.locator)
        if not locator:
            return False
        self.controller.write_workflow(locator)

    def save_log(self, locator_class, force_choose_locator=True):
        self.flush_changes()
        gui_get = locator_class.save_from_gui
        if force_choose_locator:
            locator = gui_get(self, Log.vtType,
                              self.controller.locator)
        else:
            locator = (self.controller.locator or
                       gui_get(self, Log.vtType,
                               self.controller.locator))
        if locator is not None and locator.is_untitled():
            locator = gui_get(self, Log.vtType,
                              self.controller.locator)
        if not locator:
            return False
        self.controller.write_log(locator)

    def save_registry(self, locator_class, force_choose_locator=True):
        self.flush_changes()
        gui_get = locator_class.save_from_gui
        if force_choose_locator:
            locator = gui_get(self, ModuleRegistry.vtType,
                              self.controller.locator)
        else:
            locator = (self.controller.locator or
                       gui_get(self, ModuleRegistry.vtType,
                               self.controller.locator))
        if locator is not None and locator.is_untitled():
            locator = gui_get(self, ModuleRegistry.vtType,
                              self.controller.locator)
        if not locator:
            return False
        self.controller.write_registry(locator)

    def save_version_graph(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self.window(),
            "Save DOT...",
            vistrails_file_directory(),
            "Graphviz DOT files (*.dot)")[0]
        if not filename:
            return
        self.controller.save_version_graph(filename)


    def save_opm(self, locator_class=XMLFileLocator, 
             force_choose_locator=True):
        self.flush_changes()
        gui_get = locator_class.save_from_gui
        if force_choose_locator:
            locator = gui_get(self, OpmGraph.vtType,
                              self.controller.locator)
        else:
            locator = (self.controller.locator or
                       gui_get(self, OpmGraph.vtType,
                               self.controller.locator))
        if locator is not None and locator.is_untitled():
            locator = gui_get(self, OpmGraph.vtType,
                              self.controller.locator)
        if not locator:
            return False
        self.controller.write_opm(locator)
        
    
    def save_prov(self, locator_class=XMLFileLocator, 
             force_choose_locator=True):
        self.flush_changes()
        gui_get = locator_class.save_from_gui
        if force_choose_locator:
            locator = gui_get(self, ProvDocument.vtType,
                              self.controller.locator)
        else:
            locator = (self.controller.locator or
                       gui_get(self, ProvDocument.vtType,
                               self.controller.locator))
        if locator is not None and locator.is_untitled():
            locator = gui_get(self, ProvDocument.vtType,
                              self.controller.locator)
        if not locator:
            return False
        self.controller.write_prov(locator)


    def has_changes(self):
        return self.controller.changed

    def flush_changes(self):
        """Flush changes in the vistrail before closing or saving.
        """
        # Quick workaround for notes focus out bug (ticket #182)
        # There's probably a much better way to fix this.
        from vistrails.gui.version_prop import QVersionProp
        prop = QVersionProp.instance()
        prop.versionNotes.commit_changes()

    def execute(self):
        # makes sure we are not already executing
        if self.is_executing:
            return
        self.is_executing = True

        view = self.get_current_tab()
        try:
            if hasattr(view, 'execute'):
                view.setFocus(QtCore.Qt.MouseFocusReason)
                view.execute()
        finally:
            self.is_executing = False

    def publish_to_web(self):
        view = self.get_current_tab()
        if hasattr(view, 'publish_to_web'):
            view.publish_to_web()
                 
    def publish_to_paper(self):
        view = self.get_current_tab()
        if hasattr(view, 'publish_to_paper'):
            view.publish_to_paper()

    def get_mashup_from_mashuptrail_id(self, mashuptrail_id, mashupVersion):
        """get_mashup_from_mashuptrail_id(mashuptrail_id: int,
                                           mashupVersion: int/str) -> None
        It will find the matching mashuptrail and return the mashup
        mashupVersion can be either version number or version tag

        """
        for mashuptrail in self.controller._mashups:
            if str(mashuptrail.id) == mashuptrail_id:
                try:
                    mashupVersion = int(mashupVersion)
                except ValueError:
                    mashupVersion = mashuptrail.getTagMap()[mashupVersion]
                mashup = mashuptrail.getMashup(mashupVersion)
                return mashup
        return None

    def open_mashup(self, mashup):
        """open_mashup(mashup: Mashup) -> None
        It will switch to version view, select the corresponding node 
        and run the mashup """
        from vistrails.gui.version_prop import QVersionProp
        #first we will show the hisotry view and select the version that has
        #this mashup
        vt_version = mashup.version
        window = self.window()
        window.qactions['history'].trigger()
        self.version_selected(vt_version, by_click=True)
        self.version_view.select_current_version()
        #then we will execute the mashup
        version_prop = QVersionProp.instance()
        version_prop.versionMashups.openMashup(mashup.id)
        
    def edit_mashup(self, mashup):
        """edit_mashup(mashup: Mashup) -> None
        It will select the corresponding node, switch to mashup view, 
        and select mashup """
        from vistrails.gui.mashups.mashups_inspector import QMashupsInspector
        vt_version = mashup.version
        window = self.window()
        window.qactions['history'].trigger()
        self.version_selected(vt_version, by_click=True)
        self.version_view.select_current_version()
        window.qactions['mashup'].trigger()
        inspector = QMashupsInspector.instance()
        inspector.mashupsList.selectMashup(mashup.name)
        
    def open_parameter_exploration(self, pe_id):
        pe = self.controller.vistrail.db_get_parameter_exploration_by_id(pe_id)
        if not pe:
            return
        if self.controller.current_version != pe.action_id:
            self.window().qactions['history'].trigger()
            self.version_selected(pe.action_id, True)
            self.version_view.select_current_version()
        self.controller.current_parameter_exploration = pe
        self.pe_view.setParameterExploration(pe)
        self.window().qactions['explore'].trigger()

    def apply_parameter_exploration(self, version, pe):
        if not pe:
            return
        pe = pe.__copy__()
        pe.action_id = version
        if self.controller.current_version != version:
            self.window().qactions['history'].trigger()
            self.version_selected(version, True)
            self.version_view.select_current_version()
        #self.controller.current_parameter_exploration = pe
        self.window().qactions['explore'].trigger()
        self.pe_view.setParameterExploration(pe, False)
        self.pe_view.table.setPipeline(self.controller.current_pipeline)

    ##########################################################################
    # Undo/redo
        
    def set_pipeline_selection(self, old_action, new_action):
        # need to check if anything on module changed or
        # any connections changed
        module_types = set(['module', 'group', 'abstraction'])
        module_child_types = set(['function', 'parameter', 'location', 
                                  'portSpec', 'annotation'])
        conn_types = set(['connection'])
        conn_child_types = set(['port'])

        view = self.stack.currentWidget()
        if not isinstance(view, QPipelineView):
            return 
        
        pipeline_scene = view.scene()    

        if old_action is None:
            old_action_id = 0
        else:
            old_action_id = old_action.id
        if new_action is None:
            new_action_id = 0
        else:
            new_action_id = new_action.id
        action = self.controller.vistrail.general_action_chain(old_action_id,
                                                               new_action_id)

        def module_change():
            module_ids = set()
            function_ids = set()
            for op in action.operations:
                if op.what in module_types and \
                        (op.vtType == 'change' or op.vtType == 'add'):
                    module_ids.add(op.objectId)
                elif op.what in module_child_types and \
                        (op.vtType == 'change' or op.vtType == 'add' or
                         op.vtType == 'delete'):
                    if op.what == 'parameter':
                        function_ids.add(op.parentObjId)
                    else:
                        module_ids.add(op.parentObjId)
            if len(function_ids) > 0:
                for m_id, module in \
                        self.controller.current_pipeline.modules.items():
                    to_discard = set()
                    for f_id in function_ids:
                        if module.has_function_with_real_id(f_id):
                            module_ids.add(m_id)
                            to_discard.add(f_id)
                    function_ids -= to_discard

            for id in module_ids:
                if id in pipeline_scene.modules:
                    pipeline_scene.modules[id].setSelected(True)

        def connection_change():
            conn_ids = set()
            for op in action.operations:
                if op.what in conn_types and \
                        (op.vtType == 'change' or op.vtType == 'add'):
                    conn_ids.add(op.objectId)
                elif op.what in conn_child_types and \
                        (op.vtType == 'change' or op.vtType == 'add' or 
                         op.vtType == 'delete'):
                    conn_ids.add(op.parentObjId)
            for id in conn_ids:
                if id in pipeline_scene.connections:
                    pipeline_scene.connections[id].setSelected(True)
                    
        module_change()
        connection_change()
        
    def undo(self):
        (old_action, new_action) = self.controller.undo()
        self.set_pipeline_selection(old_action, new_action)
        if new_action is not None:
            return new_action.id
        return 0
    
    def redo(self):
        (old_action, new_action) = self.controller.redo()
        self.set_pipeline_selection(old_action, new_action)
        return new_action.id

    def setVistrailVarsMode(self, on):
        """ setVistrailVarsMode(on: bool) -> None
        Set the vistrail variable panel state for the view

        """
        if on:
            self.pipelineTab.vistrailVars.toolWindow().show()
        else:
            self.pipelineTab.vistrailVars.toolWindow().hide()

    def stateChanged(self):
        """ stateChanged() -> None
        Handles 'stateChanged' signal from VistrailController """
        from vistrails.gui.vistrails_window import _app
        _app.notify("state_changed", self)
        _app.state_changed(self)
        
################################################################################
# FIXME: There is a bug on VisTrails that shows up if you load terminator.vt,
# open the image slices HW, undo about 300 times and then try to redo.
# This should be a test here, as soon as we have an api for that.

if __name__=="__main__":
    # Initialize the Vistrails Application and Theme
    import sys
    from vistrails.gui import qt, theme
    app = qt.createBogusQtGuiApp(sys.argv)
    theme.initializeCurrentTheme()

    # Now visually test QPipelineView
    vv = QVistrailView(None)
    vv.show()    
    sys.exit(app.exec_())





import contextlib
import io
import logging
import os
import shutil
import urllib.request, urllib.parse, urllib.error
import sys

from vistrails.core import debug
from vistrails.core.bundles import py_import
from vistrails.core.modules.basic_modules import PathObject
from vistrails.core.modules.config import ModuleSettings
from vistrails.core.modules.vistrails_module import Module, ModuleError, \
    ModuleSuspended
from vistrails.core.vistrail.job import JobMixin


tej = py_import('tej', {'pip': 'tej'})


assert __name__.endswith('.init')
this_pkg = __name__[:-5]


class ServerLogger(tej.ServerLogger):
    """Subclass of tej server logger that can be toggled on and off.

    This is used to hide server messages from the VisTrails console when
    running tej commands.
    """
    def __init__(self):
        self.hidden = False
        tej.ServerLogger.__init__(self)

    @contextlib.contextmanager
    def hide_output(self):
        self.hidden = True
        try:
            yield
        finally:
            self.hidden = False

    def message(self, data):
        if self.hidden:
            debug.debug("tej server: %s" % data)
        else:
            debug.warning("tej server: %s" % data)

ServerLogger = ServerLogger()


class RemoteQueue(tej.RemoteQueue):
    """Subclass of tej's RemoteQueue that uses our `ServerLogger`.
    """
    def server_logger(self):
        return ServerLogger


class QueueCache(object):
    """A global cache of RemoteQueue objects.
    """
    def __init__(self):
        self._cache = {}

    def get(self, destination, queue, setup_runtime, need_runtime):
        key = (destination,
               queue,
               setup_runtime or None,
               tuple(sorted(need_runtime)) if need_runtime else None)
        if key in self._cache:
            return self._cache[key]
        else:
            queue = RemoteQueue(destination, queue,
                                setup_runtime=setup_runtime,
                                need_runtime=need_runtime)
            self._cache[key] = queue
            return queue

QueueCache = QueueCache()


class Queue(Module):
    """A connection to a queue on a remote server.

    `hostname` can be a hostname or a full destination in the format:
    ``[ssh://][user@]server[:port]``, e.g. ``vistrails@nyu.edu``.

    If `setup_runtime` is set, it should be the name of the runtime to deploy
    on the server if the queue doesn't already exist. If None (the default),
    tej will select what is appropriate automatically. If `need_runtime` is
    set, this should be one of the accepted values.

    If `need_runtime` is set, it is a comma-separated list of runtime names
    that are acceptable. If the queue already exists on the server and this
    argument is not None, the installed runtime will be matched against it, and
    execution will fail if it is not one of the provided values.
    """
    _input_ports = [('hostname', '(basic:String)'),
                    ('username', '(basic:String)',
                     {'optional': True}),
                    ('port', '(basic:Integer)',
                     {'optional': True, 'defaults': "['22']"}),
                    ('queue', '(basic:String)',
                     {'optional': True, 'defaults': "['~/.tej']"}),
                    ('setup_runtime', '(basic:String)',
                     {'optional': True}),
                    ('need_runtime', '(basic:String)',
                     {'optional': True})]
    _output_ports = [('queue', '(org.vistrails.extra.tej:Queue)')]

    def compute(self):
        destination_str = self.get_input('hostname')
        if self.has_input('username') or self.has_input('port'):
            destination = {'hostname': destination_str,
                           'username': self.get_input('username'),
                           'port': self.get_input('port')}
            destination_str = tej.destination_as_string(destination)

        if self.has_input('setup_runtime'):
            setup_runtime = self.get_input('setup_runtime')
        else:
            setup_runtime = None

        if self.has_input('need_runtime'):
            need_runtime = set(
                e.strip()
                for e in self.get_input('need_runtime').split(','))
        else:
            need_runtime = None

        queue = self.get_input('queue')
        self.set_output('queue', QueueCache.get(destination_str, queue,
                                                setup_runtime, need_runtime))


class AssembleDirectoryMixin(object):
    """A mixin for assembling a directory from paths passed on port specs.

    Modules using this mixin should use the DirectoryConfigurationWidget
    configuration widget to setup the input port specs.
    :meth:`assemble_directory` returns a PathObject for the temporary directory
    with the paths copied into it.
    """
    def __init__(self):
        self.input_ports_order = []

    def transfer_attrs(self, module):
        super(AssembleDirectoryMixin, self).transfer_attrs(module)
        self.input_ports_order = [p.name for p in module.input_port_specs]

    def assemble_directory(self, base=None, do_copy=False):
        """Create the directory using the optional `base` and the port specs.

        :type base: PathObject

        This creates a directory that contains all the contents of `base` (if
        provided, else start empty), and all the additional files and
        directories provided as input port specs (configured using the
        `DirectoryConfigurationWidget`).

        If there are no input port specs and `do_copy` is False, this method
        will just return the currect location of `base` without copying it to a
        temporary directory.
        """
        if self.input_ports_order or do_copy:
            directory = self.interpreter.filePool.create_directory(
                    prefix='vt_tmp_makedirectory_')

            # Copy everything in base
            if base is not None:
                for name in os.listdir(base.name):
                    src = os.path.join(base.name, name)
                    dst = os.path.join(directory.name, name)
                    if os.path.isdir(src):
                        shutil.copytree(src, dst)
                    else:
                        shutil.copy2(src, dst)
            # Copy from port specs
            for name in self.input_ports_order:
                shutil.copy(self.get_input(name).name,
                            os.path.join(directory.name, name))
            return directory
        else:
            return base


class MakeDirectory(AssembleDirectoryMixin, Module):
    """Creates a temporary directory and puts the given files in it.
    """
    _settings = ModuleSettings(configure_widget=(
            '%s.widgets' % this_pkg, 'DirectoryConfigurationWidget'))
    _output_ports = [('directory', '(basic:Directory)')]

    def __init__(self):
        AssembleDirectoryMixin.__init__(self)
        Module.__init__(self)

    def compute(self):
        directory = self.assemble_directory()
        self.set_output('directory', directory)


class RemoteJob(object):
    """This implements the JobMonitor's JobHandle interface.

    These objects are returned to the JobMonitor via the ModuleSuspended
    exceptions; they are used by the JobMonitor to figure out the status of the
    submitted job.
    """
    def __init__(self, queue, job_id):
        self.queue = queue
        self.job_id = job_id

    def finished(self):
        try:
            with ServerLogger.hide_output():
                status, target, arg = self.queue.status(self.job_id)
        except tej.JobNotFound:
            # We signal that we are done
            # The Module will raise an error on resume, as intended
            return True
        return status == tej.RemoteQueue.JOB_DONE


class Job(Module):
    """A reference to a job in a queue.

    Objects represented by this type only represent completed jobs, since else,
    the creating module would have failed/suspended.

    You probably won't use this module directly since it references a
    pre-existing job by name; just use one of the SubmitJob modules that
    compute the upstream's signature.
    """
    _input_ports = [('id', '(basic:String)'),
                    ('queue', Queue)]
    _output_ports = [('job', '(org.vistrails.extra.tej:Job)'),
                     ('exitcode', '(basic:Integer)')]

    def compute(self):
        queue = self.get_input('queue')
        job_id = self.get_input('id')

        # Check job status
        try:
            with ServerLogger.hide_output():
                status, target, arg = queue.status(job_id)
        except tej.JobNotFound:
            raise ModuleError(self, "Job not found")

        # Create job object
        job = RemoteJob(queue=queue, job_id=job_id)

        if status == tej.RemoteQueue.JOB_DONE:
            self.set_output('job', job)
            self.set_output('exitcode', int(arg))
        elif status == tej.RemoteQueue.JOB_RUNNING:
            raise ModuleSuspended(self, "Remote job is running",
                                  handle=job)
        else:
            raise ModuleError(self, "Invalid job status %r" % status)


class BaseSubmitJob(JobMixin, Module):
    """Starts a job on a server.

    Thanks to the suspension/job tracking mechanism, this module does much more
    than start a job. If the job is running, it will suspend again. If the job
    is finished, you can obtain files from it.
    """
    _settings = ModuleSettings(abstract=True)
    _input_ports = [('queue', Queue),
                    ('print_error', '(basic:Boolean)',
                     {'optional': True, 'defaults': "['False']"})]
    _output_ports = [('job', '(org.vistrails.extra.tej:Job)'),
                     ('exitcode', '(basic:Integer)')]

    def make_id(self):
        """Makes a default identifier, using the pipeline signature.

        Unused if we have an explicit identifier.
        """
        if not hasattr(self, 'signature'):
            raise ModuleError(self,
                              "No explicit job ID and module has no signature")
        return "vistrails_module_%s" % self.signature

    def job_read_inputs(self):
        """Reads the input ports.
        """
        queue = self.get_input('queue')
        return {'destination': queue.destination_string,
                'queue': str(queue.queue),
                'setup_runtime': queue.setup_runtime or '',
                'need_runtime': (','.join(queue.need_runtime)
                                 if queue.need_runtime is not None else ''),
                'job_id': self.make_id()}

    def job_start(self, params):
        """Submits a job.

        Reimplement in subclasses to actually submit a job.
        """
        raise NotImplementedError

    def job_get_handle(self, params):
        """Gets a RemoteJob object to monitor a runnning job.
        """
        queue = QueueCache.get(params['destination'], params['queue'],
                               params.get('setup_runtime') or None,
                               params.get('need_runtime') or None)
        return RemoteJob(queue, params['job_id'])

    def job_finish(self, params):
        """Finishes job.

        Gets the exit code from the server.
        """
        queue = QueueCache.get(params['destination'], params['queue'],
                               params.get('setup_runtime') or None,
                               params.get('need_runtime') or None)
        with ServerLogger.hide_output():
            status, target, arg = queue.status(params['job_id'])
        assert status == tej.RemoteQueue.JOB_DONE
        params['exitcode'] = int(arg)
        return params

    def job_set_results(self, params):
        """Sets the output ports once the job is finished.
        """
        queue = QueueCache.get(params['destination'], params['queue'],
                               params.get('setup_runtime') or None,
                               params.get('need_runtime') or None)
        self.set_output('exitcode', params['exitcode'])
        self.set_output('job', RemoteJob(queue, params['job_id']))

        if self.get_input('print_error') and params['exitcode'] != 0:
            # If it failed and 'print_error' is set, download and print stderr
            # then fail
            destination = self.interpreter.filePool.create_file(
                    prefix='vt_tmp_job_err_').name
            with ServerLogger.hide_output():
                queue.download(params['job_id'],
                               '_stderr',
                               destination=destination,
                               recursive=False)
            with open(destination, 'r') as fin:
                chunk = fin.read(4096)
                sys.stderr.write(chunk)
                while len(chunk) == 4096:
                    chunk = fin.read(4096)
                    if chunk:
                        sys.stderr.write(chunk)
            raise ModuleError(self,
                              "Job failed with status %d" % params['exitcode'])


class SubmitJob(AssembleDirectoryMixin, BaseSubmitJob):
    """Submits a generic job (a directory).
    """
    _settings = ModuleSettings(configure_widget=(
            '%s.widgets' % this_pkg, 'DirectoryConfigurationWidget'))
    _input_ports = [('job', '(basic:Directory)',
                     {'optional': True}),
                    ('script', '(basic:String)',
                     {'optional': True, 'defaults': "['start.sh']"})]

    def __init__(self):
        AssembleDirectoryMixin.__init__(self)
        Module.__init__(self)

    def job_start(self, params):
        """Sends the directory and submits the job.
        """
        queue = QueueCache.get(params['destination'], params['queue'],
                               params.get('setup_runtime') or None,
                               params.get('need_runtime') or None)

        # First, check if job already exists
        try:
            with ServerLogger.hide_output():
                queue.status(params['job_id'])
        except (tej.JobNotFound, tej.QueueDoesntExist):
            pass
        else:
            return params

        if self.has_input('job'):
            job_dir = self.get_input('job')
            if not os.path.exists(job_dir.name):
                raise ModuleError(self, "Directory doesn't exist")
        else:
            job_dir = None

        # Use AssembleDirectoryMixin to get additional files from port specs
        job_dir = self.assemble_directory(job_dir, False)

        # Check that the script exists
        script = self.get_input('script')
        if not os.path.exists(os.path.join(job_dir.name, script)):
            raise ModuleError(self, "Script does not exist")

        # Alright, submit a new job
        queue.submit(params['job_id'],
                     job_dir.name,
                     script)
        return params


class SubmitShellJob(BaseSubmitJob):
    """Submits a shell script.
    """
    _settings = ModuleSettings(configure_widget=(
            '%s.widgets' % this_pkg, 'ShellSourceConfigurationWidget'))
    _input_ports = [('source', '(basic:String)')]
    _output_ports = [('stderr', '(basic:File)'),
                     ('stdout', '(basic:File)')]

    def job_start(self, params):
        """Creates a temporary job with the given source, upload and submit it.
        """
        queue = QueueCache.get(params['destination'], params['queue'],
                               params.get('setup_runtime') or None,
                               params.get('need_runtime') or None)

        # First, check if job already exists
        try:
            with ServerLogger.hide_output():
                queue.status(params['job_id'])
        except (tej.JobNotFound, tej.QueueDoesntExist):
            pass
        else:
            return params

        # Alright, submit a new job
        directory = self.interpreter.filePool.create_directory(
                prefix='vt_tmp_shelljob_').name
        # We use io.open() here because we could be writing scripts on Windows
        # before uploading them to a POSIX server
        source = urllib.parse.unquote(self.get_input('source'))
        if isinstance(source, bytes):
            kwargs = {'mode': 'wb'}
        else:
            kwargs = {'mode': 'w', 'newline': '\n'}
        with io.open(os.path.join(directory, 'start.sh'),
                     **kwargs) as fp:
            fp.write(source)

        queue.submit(params['job_id'], directory)

        return params

    def job_set_results(self, params):
        """Gets stderr and stdout.
        """
        super(SubmitShellJob, self).job_set_results(params)

        temp_dir = self.interpreter.filePool.create_directory(
                prefix='vt_tmp_shelljobout_').name
        queue = QueueCache.get(params['destination'], params['queue'],
                               params.get('setup_runtime') or None,
                               params.get('need_runtime') or None)
        queue.download(params['job_id'], ['_stderr', '_stdout'],
                       directory=temp_dir)
        self.set_output('stderr',
                        PathObject(os.path.join(temp_dir, '_stderr')))
        self.set_output('stdout',
                        PathObject(os.path.join(temp_dir, '_stdout')))


class DownloadFile(Module):
    """Downloads a file from a remote job.
    """
    _input_ports = [('job', Job),
                    ('filename', '(basic:String)')]
    _output_ports = [('file', '(basic:File)')]

    def compute(self):
        job = self.get_input('job')
        assert isinstance(job, RemoteJob)

        destination = self.interpreter.filePool.create_file(
                prefix='vt_tmp_shelljobout_')
        with ServerLogger.hide_output():
            job.queue.download(job.job_id,
                               self.get_input('filename'),
                               destination=destination.name,
                               recursive=False)

        self.set_output('file', destination)


class DownloadDirectory(Module):
    """Downloads a directory from a remote job.
    """
    _input_ports = [('job', Job),
                    ('pathname', '(basic:String)')]
    _output_ports = [('directory', '(basic:Directory)')]

    def compute(self):
        job = self.get_input('job')
        assert isinstance(job, RemoteJob)

        destination = self.interpreter.filePool.create_directory(
                prefix='vt_tmp_shelljobout_').name
        target = os.path.join(destination, 'dir')
        with ServerLogger.hide_output():
            job.queue.download(job.job_id,
                               self.get_input('pathname'),
                               destination=target,
                               recursive=True)

        self.set_output('directory', PathObject(target))


_tej_log_handler = None


class _VisTrailsTejLogHandler(logging.Handler):
    """Custom handler for the 'tej' logger that re-logs to VisTrails.
    """
    def emit(self, record):
        msg = "tej: %s" % self.format(record)
        if record.levelno >= logging.CRITICAL:
            debug.critical(msg)
        elif record.levelno >= logging.WARNING:
            debug.warning(msg)
        elif record.levelno >= logging.INFO:
            debug.log(msg)
        else:
            debug.debug(msg)


def initialize():
    # Route tej log messages to VisTrails
    global _tej_log_handler
    _tej_log_handler = _VisTrailsTejLogHandler()
    tej_logger = logging.getLogger('tej')
    tej_logger.propagate = False
    tej_logger.addHandler(_tej_log_handler)
    tej_logger.setLevel(logging.DEBUG)


def finalize():
    # Unregister logging handler; useful for reloading support
    global _tej_log_handler
    tej_logger = logging.getLogger('tej')
    if _tej_log_handler is not None:
        tej_logger.removeHandler(_tej_log_handler)
        _tej_log_handler = None
    tej_logger.propagate = True


_modules = [Queue,
            Job, BaseSubmitJob, SubmitJob, SubmitShellJob,
            DownloadFile, DownloadDirectory,
            MakeDirectory]
